/*******************************************************************************
 * Copyright (c) 2005 INRIA.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Marcos Didonet Del Fabro - initial API and implementation
 *    Guillaume Gueltas
 *******************************************************************************/


package org.eclipse.gmt.weaver.extension.providers;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.common.notify.AdapterFactory;
import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.impl.EEnumImpl;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.util.EcoreEList;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.emf.ecore.util.ExtendedMetaData;
import org.eclipse.emf.ecore.util.FeatureMap;
import org.eclipse.emf.ecore.xmi.XMIResource;
import org.eclipse.emf.edit.EMFEditPlugin;
import org.eclipse.emf.edit.command.CommandParameter;
import org.eclipse.emf.edit.provider.ComposeableAdapterFactory;
import org.eclipse.emf.edit.provider.IEditingDomainItemProvider;
import org.eclipse.emf.edit.provider.IItemLabelProvider;
import org.eclipse.emf.edit.provider.IItemPropertyDescriptor;
import org.eclipse.emf.edit.provider.IItemPropertySource;
import org.eclipse.emf.edit.provider.IStructuredItemContentProvider;
import org.eclipse.emf.edit.provider.ITreeItemContentProvider;
import org.eclipse.emf.edit.provider.ItemPropertyDescriptor;
import org.eclipse.emf.edit.provider.ItemPropertyDescriptorDecorator;
import org.eclipse.emf.edit.provider.ReflectiveItemProvider;
import org.eclipse.gmt.weaver.core.WeaverModelManager;
import org.eclipse.gmt.weaver.editors.interfaces.IWeaverItemProvider;
import org.eclipse.gmt.weaver.extension.plugin.WeaverBaseExtensionPlugin;
import org.eclipse.gmt.weaver.plugin.util.WeaverInterfaceUtil;
import org.eclipse.gmt.weaver.plugin.util.WeaverResourceUtil;
import org.eclipse.gmt.weaver.plugin.util.WeaverUtil;
import org.eclipse.gmt.weaver.providers.IIdentifierAdapter;


/**
 * ItemProvider base extension
 * @since 1.0
 */

public class WElementItemProvider extends ReflectiveItemProvider 
	implements      
		IEditingDomainItemProvider,
		IStructuredItemContentProvider,
		ITreeItemContentProvider,
		IItemLabelProvider,
		IItemPropertySource, 
		IWeaverItemProvider
		
    {

	private Map icons = null;
	
    /**
     * This creates a base extension item provider
     */
    public WElementItemProvider(){
        super(null);
    }
    
    /* (non-Javadoc)
     * @see org.eclipse.gmt.weaver.editors.IWeaverItemProvider#setAdapterFactory(org.eclipse.emf.common.notify.AdapterFactory)
     */
    public void setAdapterFactory(AdapterFactory adapterFactory){
        this.adapterFactory = adapterFactory;
    }
    
	/* (non-Javadoc)
	 * @see org.eclipse.gmt.weaver.editors.IWeaverItemProvider#setIcon(java.lang.String)
	 */
	public void setIconMap(Map icons){
		this.icons = icons;
	}
	
	/**
	 * @return map of icons
	 */
	public Map getIconMap(){
		return icons;
	}
	
    /**
     * Returns the icon associed to element or null if it doesn't exist
     * @return
     */
	public String getIconPath(String element){
		try{
			Object obj = icons.get(element);
			return (String)obj;
		}catch(Exception e){
			return null;
		}
	}
	
    /* (non-Javadoc)
     * @see org.eclipse.emf.edit.provider.ItemProviderAdapter#getImage(java.lang.Object)
     */
    public Object getImage(Object object){
    	EObject eObject = (EObject)object;
    	EClass eClass = eObject.eClass();
    	String name = eClass.getName();
    	String icon = getIconPath(name);
    	if (icon != null){
    		return URI.createURI(WeaverBaseExtensionPlugin.getPlugin().getImage(icon).toString() + "#" + eClass.getName());
    	}
    	else
    		return URI.createURI(getResourceLocator().getImage("full/obj16/Item").toString() + "#" + eClass.getName());
    }
    
    /* (non-Javadoc)
     * @see org.eclipse.emf.edit.command.CreateChildCommand$Helper#getCreateChildImage(java.lang.Object, java.lang.Object, java.lang.Object, java.util.Collection)
     */
    public Object getCreateChildImage(Object owner, Object feature, Object child, Collection selection){ 
    	EObject eObject = (EObject)child;
    	EClass eClass = eObject.eClass();
    	String name = eClass.getName();
    	String icon = getIconPath(name);    	
    	
    	if (icon != null){
    		return URI.createURI(WeaverBaseExtensionPlugin.getPlugin().getImage(icon).toString());
    	}
    	else
    		return EMFEditPlugin.INSTANCE.getImage("full/ctool16/CreateChild");
    }
	
    /**
     * sets constraints on the returned children objects and adds then into a deleteList
     * @param parent the current parent obj
     * @param child the child
     * @param afectedObj an associated object (usually is the child itself) 
     * @param deleteList the constrained object that will be deleted
     */
    protected void setChildrenConstraints(EObject parent, EObject child, Object afectedObj, Collection deleteList){
        //models cant have links and association ends as child
        if (  isInstanceOf(parent,"WModel") && (
                	isInstanceOf(child,"WLinkEnd")
                    || isInstanceOf(child,"WAssociationEnd") ) )
             deleteList.add(afectedObj);
        
        //ref and model elements cannot be created by menu
        if (isInstanceOf(child,"WModelRef") || 
                isInstanceOf(child,"WElementRef")
                       || isInstanceOf(child,"WRef")
                       || isInstanceOf(child,"WModel"))            
            deleteList.add(afectedObj);                        
    }

    /* (non-Javadoc)
     * @see org.eclipse.emf.edit.provider.ItemProviderAdapter#getChildren(java.lang.Object)
     */
    public Collection getChildren(Object object){
        Collection children = super.getChildren(object);
        Collection deleteList = new ArrayList();        
        for (Iterator i = children.iterator();i.hasNext();){
            EObject eChildObj = (EObject)i.next();
            setChildrenConstraints((EObject) object, eChildObj, eChildObj, deleteList);            
        }
        children.removeAll(deleteList);
        return children;
    }

    /* (non-Javadoc)
     * @see org.eclipse.emf.edit.provider.ItemProviderAdapter#collectNewChildDescriptors(java.util.Collection, java.lang.Object)
     */
    protected void collectNewChildDescriptors(Collection newChildDescriptors, Object object)    {       
      super.collectNewChildDescriptors(newChildDescriptors, object);
      Collection deleteList = new ArrayList();
      for (Iterator i = newChildDescriptors.iterator(); i.hasNext();){
          CommandParameter commandMenuObj = (CommandParameter)i.next();
          EObject menuObj = (EObject)commandMenuObj.getValue();
          setChildrenConstraints((EObject) object, menuObj, commandMenuObj, deleteList);                    
      }
      newChildDescriptors.removeAll(deleteList);
    }    
    /**
     * returns true if its factory is of type WElementProviderAdapterFactory 
     * @param type
     * @return
     */
    public boolean isAdapterForType(Object type){
      return  type instanceof org.eclipse.gmt.weaver.providers.WeaverAdapterFactory;
    }  
    /*
     * wrapper for calling WeaverUtil.isInstanceOf
     */
    protected boolean isInstanceOf(EObject eCurrentObj, String classifierName){
        return WeaverUtil.isInstanceOf(eCurrentObj,classifierName,WeaverUtil.getMetamodelPackages(eCurrentObj));
    }
    /**
     * Returns the container resource for this eClass  
     * @param eClass
     * @return
     */
    private Resource getResource (EClass eClass){
        EPackage pkg = (EPackage)EcoreUtil.getRootContainer(eClass);
        return pkg.eResource();                
    }    
    /**
     * Gets all concrete subclasses of a class in all the packages of a resourceSet.
     * @param eClass
     * @return
     */
    protected List getAllConcreteSubclasses(EClass eClass){            
        allRoots = new ArrayList();
        allEClasses = new ArrayList();
                
        allRoots = WeaverUtil.getAllPackages(getResource(eClass));                          
        //get all packages
        for (Iterator pkgs = allRoots.iterator(); pkgs.hasNext();){
            EPackage apkg = (EPackage)pkgs.next();
            //get packages elements
            for (Iterator i = apkg.eAllContents() ;i.hasNext();){
                EObject eObj = (EObject)i.next();
                if (eObj instanceof EClass){//gets concrete and child elements
                	EClass aClass = (EClass) eObj;
                    if ( !aClass.isAbstract() && eClass.isSuperTypeOf(aClass) && !allEClasses.contains(aClass))                        
                        allEClasses.add((EClass) eObj);                  
                }
            }            
        }
        return allEClasses;
    }

    
    /**
     * returns the descriptors for all properties of a given object. It can modified to create a different 
     * editor for each property as needed
     * @param object 
     * @return
     */
   
    public List getPropertyDescriptors(Object object){
        return getEElementPropertyDescriptors(object);
    }
    
    /**
     * acts like a filter to verify if a specified feature will be shown in the property editor 
     * @param object
     * @param feature
     * @return 
     */
    private boolean isValidStructuralFeature(Object object, EStructuralFeature feature){
    	if (object instanceof EObject){
    		EObject eObj = (EObject) object;
    		//does not show the Model property for linkEnds and associationEnds
    		 if (isInstanceOf(eObj,"WLinkEnd") || isInstanceOf(eObj,"WAssociationEnd")){
    		 	return ! feature.getName().equals("model");     		 	
    		 }
    	}    	
    	return true;
    }
    
    /**
     * gets the properties of a gives object. Does a loop to get all structural features
     * @param object
     * @return a list
     */
    public List getEElementPropertyDescriptors(Object object)     {
        List descriptors = new ArrayList();        
        for (Iterator i = ((EObject)object).eClass().getEAllStructuralFeatures().iterator(); i.hasNext(); ){
            EStructuralFeature eFeature = (EStructuralFeature)i.next();
            if (isValidStructuralFeature(object, eFeature)){
                //flat properties = attributes
                if (!(eFeature instanceof EReference) || !((EReference)eFeature).isContainment()) {
                	if (eFeature instanceof EReference && ((EReference)eFeature).isMany()) {
                		descriptors.add (setReferenceMultipleCardProperty(object,eFeature));
                	} else {
                    	descriptors.add (setFlatProperty(object,eFeature));
                    	// the annotation attributes return Eobjects, but they should be treated as simple flat properties
                        if ( (((EObject)object).eGet(eFeature) instanceof EObject) && ! (eFeature.getEType() instanceof EEnumImpl))                        
                            setNestedObjectProperties((EObject)((EObject)object).eGet(eFeature),eFeature,descriptors);                		
                	}
                }     
                else{// Containment properties
                    setContainementProperty(((EObject)object).eGet(eFeature), eFeature, descriptors);
                }            	
            }
        }
        itemPropertyDescriptors = descriptors;
        return itemPropertyDescriptors;
    }
    
    /**
     * gets the properties of objects when this object is itself a property (nested properties)
     * @param eObj
     * @param eFeature
     * @param result
     * @return
     */
    private Collection setNestedObjectProperties(EObject eObj, EStructuralFeature eFeature, Collection result){       
        ItemPropertyDescriptor descr = (new ItemPropertyDescriptor
                (((ComposeableAdapterFactory)adapterFactory).getRootAdapterFactory(),
                        getFeatureText(eFeature),
                        getResourceLocator().getString
                        ("_UI_Property_description", new Object [] { getFeatureText(eFeature), eFeature.getEType().getName() }),
                        eFeature,
                        false,
                        ItemPropertyDescriptor.GENERIC_VALUE_IMAGE));
          List aList = getObjectProperties(eObj, "WElementRef",  eObj.eClass().getName(), 0);
          if (aList != null)
        	  result.addAll(aList);
           //eObj.eClass().getName(),
        return result;
    }
    /**
     * sets Containment properties
     * @param ecoreList
     * @param eFeature
     * @param result
     * @return
     */
    private ItemPropertyDescriptor setContainementProperty(Object ecoreList, EStructuralFeature eFeature, Collection result){
        //object(nested) properties
        //List result = new ArrayList();
        ItemPropertyDescriptor descr = null;
        if (ecoreList instanceof EcoreEList){
            EcoreEList list = (EcoreEList) ecoreList;                                    
            descr = (new ItemPropertyDescriptor
                    (((ComposeableAdapterFactory)adapterFactory).getRootAdapterFactory(),
                            getFeatureText(eFeature),
                            getResourceLocator().getString
                            ("_UI_Property_description", new Object [] { getFeatureText(eFeature), eFeature.getEType().getName() }),
                            eFeature,
                            false,
                            ItemPropertyDescriptor.GENERIC_VALUE_IMAGE));              
            //for all descriptors, look for WElementRef elements and sets a new category for them
            result.add(descr);
            int generatedID = 0;
            for (Iterator i = list.iterator(); i.hasNext();) {
                generatedID++;
                EObject reference = (EObject) i.next();
                result.addAll(getObjectProperties(reference,"WModelRef", reference.eClass().getName() + generatedID,generatedID + 100));
                        //reference.eClass().getName(), replace ModelRef and all containment properties will be shown
            }
        }
        return descr;       
    }
    
     
    private ItemPropertyDescriptor setReferenceMultipleCardProperty(Object object, EStructuralFeature eFeature){
        return (new ItemPropertyDescriptor
                (((ComposeableAdapterFactory)adapterFactory).getRootAdapterFactory(),
                        getFeatureText(eFeature),
                        getResourceLocator().getString
                        ("_UI_Property_description", new Object [] { getFeatureText(eFeature), eFeature.getEType().getName() }),
                        eFeature,
                        true,
                        ItemPropertyDescriptor.GENERIC_VALUE_IMAGE,"_"+((EObject)object).eClass().getName()));
            
    }
    
	/**
	 * adds a new flat property descriptor
	 * @param object
	 * @param eFeature
	 * @return
	 */    
    private ItemPropertyDescriptor setFlatProperty(Object object, EStructuralFeature eFeature){
        return (new ItemPropertyDescriptor
                (((ComposeableAdapterFactory)adapterFactory).getRootAdapterFactory(),
                        getFeatureText(eFeature),
                        getResourceLocator().getString
                        ("_UI_Property_description", new Object [] { getFeatureText(eFeature), eFeature.getEType().getName() }),
                        eFeature,
                        !eFeature.isMany(),
                        ItemPropertyDescriptor.GENERIC_VALUE_IMAGE,"_"+((EObject)object).eClass().getName()));
            
    }

    /**
     * returns the properties decorators for a given object
     * @param eObj
     * @param objName
     * @param category
     * @param id
     * @return
     */
    private List getObjectProperties(EObject eObj, String objName, String category, int id){
        List result= new ArrayList();
        if ( isInstanceOf(eObj,objName)){                    
            List elements = super.getPropertyDescriptors(eObj);//get the properties of the object
            int generatedID = id;//counter to add into the property id
            for (Iterator i = elements.iterator(); i.hasNext(); ){
                generatedID++;
                ItemPropertyDescriptor descriptor = (ItemPropertyDescriptor)i.next();
                Object decorator = new ElementDecorator(eObj, descriptor, generatedID + id, category);
                result.add(decorator);
            }
        }       
        return result;
    }

    /* (non-Javadoc)
     * @see org.eclipse.emf.common.notify.Adapter#notifyChanged(org.eclipse.emf.common.notify.Notification)
     */
    public void notifyChanged (Notification notification){    	
    	//Notification.EVENT_TYPE_COUNT is used as a minimal number for user defined events     	
    	if (notification.getEventType() < Notification.EVENT_TYPE_COUNT)
    		super.notifyChanged(notification);     	
    	//generates object ids 
    	if (notification.getNewValue() != null || notification.getEventType() > Notification.EVENT_TYPE_COUNT) { 
    		//do not call while deleting objects.  
    		if (notification.getNotifier() instanceof EObject){    			
    			EObject eObj = (EObject) notification.getNotifier();        		
    			IIdentifierAdapter adapter = WeaverInterfaceUtil.createGetIdentifierAdapter(eObj);
    			//code replicated in WeaverInitialize 
    			if (adapter.getID() == null){//sets the id as the class name + a date in milliseconds + number of objects
    				int a = 0;    				
    				XMIResource res = (XMIResource) eObj.eResource();
    				String className = eObj.eClass().getName();
    				do {
    					a ++;
    				} while (res != null && res.getEObject(className + a)!= null);    				    				
    				adapter.setID(className + a);
    			}
    		}    	    		
    	}
    }
    
    /* (non-Javadoc)
     * @see org.eclipse.emf.edit.provider.ItemProviderAdapter#getChildrenFeatures(java.lang.Object)
     */
    protected Collection getChildrenFeatures(Object object){
    	//derived 	
    	childrenFeatures = new ArrayList();
    	
    	EObject eObject = (EObject)object;
    	EClass eClass = eObject.eClass();
    	if (ExtendedMetaData.INSTANCE.getContentKind(eClass) != ExtendedMetaData.MIXED_CONTENT){
    		for (Iterator i = eClass.getEAllReferences().iterator(); i.hasNext(); ){
    			EReference eReference = (EReference)i.next();
    			if (eReference.isContainment() && ExtendedMetaData.INSTANCE.getGroup(eReference) == null){
    				if (!eReference.isDerived() ){
						childrenFeatures.add(eReference);    				    					    					
    				}
    			}
    		}
    	}
    	for (Iterator i = eClass.getEAllAttributes().iterator(); i.hasNext(); ){
    		EAttribute eAttribute = (EAttribute)i.next();
    		if (ExtendedMetaData.INSTANCE.getGroup(eAttribute) == null && 
    				eAttribute.getEType().getInstanceClass() == FeatureMap.Entry.class ){
    			childrenFeatures.add(eAttribute);
    		}
    	}    	
      return childrenFeatures;
    }
    
    /* (non-Javadoc)
     * @see org.eclipse.emf.edit.command.CreateChildCommand$Helper#getCreateChildText(java.lang.Object, java.lang.Object, java.lang.Object, java.util.Collection)
     */
    public String getCreateChildText(Object owner, Object feature, Object child, Collection selection){
    	String returnValue = "";
    	if (child instanceof EObject) {
    		returnValue =((EObject)child).eClass().getName();
    	}
    	if (feature instanceof EReference){
    		EReference ref = (EReference) feature;
    		if (ref.getEAnnotations().size() > 0 && !ref.isDerived()){
    			returnValue = "<<" +ref.getName() +">> " + returnValue;    			
    		}    
    		else
    			returnValue = "<" +ref.getName() +"> " + returnValue;
    	}    	
    	return returnValue;
    }
    
    /* (non-Javadoc)
     * @see org.eclipse.emf.edit.provider.ItemProviderAdapter#getText(java.lang.Object)
     */
    public String getText(Object object) {
    	String str = super.getText(object);
    	if (object instanceof EObject){
    		EObject eObj = (EObject) object;
    		if (eObj.eContainingFeature() != null ){
        		if ( eObj.eContainingFeature() instanceof EReference ){                	
                		EReference ref = (EReference) eObj.eContainingFeature();
                		if (ref.getEAnnotations().size() > 0 && !ref.isDerived()){
                			str = "<<" +ref.getName() +">> " + str;    			
                		}
                		else
                			str = "<" +ref.getName() +"> " + str;
                	}    	    	    			
    		}    		
    	}
    	return str;
    }
    
    /**
     * Property decorator : creates the object in a new category and with a distinct ID
     */
    class ElementDecorator extends ItemPropertyDescriptorDecorator {
        int number = 0;
        String category;
        
        /**
         * This creates a element decorator
         * @param obj
         * @param descr
         * @param number
         * @param category
         */
        public ElementDecorator(Object obj, IItemPropertyDescriptor descr, int number, String category){
            super(obj,descr);
            this.number =number;
            this.category = category;
        }
        
        /* (non-Javadoc)
         * @see org.eclipse.emf.edit.provider.IItemPropertyDescriptor#getCategory(java.lang.Object)
         */
        public String getCategory(Object obj){
            return category;
        }
        
        /* (non-Javadoc)
         * @see org.eclipse.emf.edit.provider.IItemPropertyDescriptor#getId(java.lang.Object)
         */
        public String getId(Object obj){
            return  (new Integer(number)).toString() + category;
        } 
        
        /* (non-Javadoc)
         * @see org.eclipse.emf.edit.provider.IItemPropertyDescriptor#canSetProperty(java.lang.Object)
         */
        public boolean canSetProperty(Object thisObject){
            return false;
        }        
    }
        
}



