package org.eclipse.gmt.weaver.transformation.operations;

import java.io.File;
import java.io.IOException;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.xml.parsers.ParserConfigurationException;

import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.xmi.impl.XMIResourceImpl;
import org.eclipse.gmt.weaver.core.WeaverExtensionManager;
import org.eclipse.gmt.weaver.core.WeaverXMLMetadata;
import org.eclipse.gmt.weaver.core.data.WovenModelDescriptor;
import org.eclipse.gmt.weaver.exception.WeaverException;
import org.eclipse.gmt.weaver.plugin.WeaverConstant;
import org.eclipse.gmt.weaver.plugin.util.WeaverUtil;
import org.eclipse.gmt.weaver.providers.IIdentifierAdapter;
import org.eclipse.gmt.weaver.providers.ResourceIdentifierAdapter;
import org.eclipse.m2m.atl.core.IModel;
import org.eclipse.m2m.atl.core.emf.EMFModel;
import org.eclipse.m2m.atl.engine.emfvm.lib.AbstractStackFrame;
import org.eclipse.m2m.atl.engine.emfvm.lib.ExecEnv;
import org.eclipse.m2m.atl.engine.emfvm.lib.LibExtension;
import org.eclipse.m2m.atl.engine.emfvm.lib.OclUndefined;
import org.eclipse.m2m.atl.engine.emfvm.lib.Operation;
import org.xml.sax.SAXException;

import uk.ac.shef.wit.simmetrics.similaritymetrics.Levenshtein;

public class UtilOperation implements LibExtension {

	@Override
	public void apply(ExecEnv execEnv, Map<String, Object> options) {
		execEnv.registerOperation(Object.class, getReferredElement());
		execEnv.registerOperation(Object.class, generateModelRef());
		execEnv.registerOperation(Object.class, getElementID());
		execEnv.registerOperation(Object.class, getElementIDbyRefType());
		execEnv.registerOperation(Object.class, setElementIDbyRefType());
		execEnv.registerOperation(Object.class, getStringSim());
	}
	
	private Operation setElementIDbyRefType() {
		return new Operation (3,"setElementIDbyRefType"){
			@Override
			public Object exec(AbstractStackFrame frame) {
				Object[] localVars = frame.getLocalVars();
				EObject eObj= null;
				String ret=null;
				String elementRef = null, ID = null;
				if (localVars[0] instanceof EObject && localVars[1] instanceof String && localVars[2] instanceof String) {
					eObj = (EObject) localVars[0];
					elementRef = (String) localVars[1];
					ID = (String) localVars[2];
				}
				else
					return null;
				WeaverExtensionManager em = new WeaverExtensionManager(null);
				IIdentifierAdapter adapter = null;
				Object aclass = null;
				try {
					aclass = em.getProviderImplementationClassByType(elementRef);
				} catch (WeaverException e) {
					e.printStackTrace();
				}
				if (aclass instanceof IIdentifierAdapter)  {
					adapter = (IIdentifierAdapter) aclass;
					adapter.setTarget(eObj);
					// TODO: check this if condition. it was eObj == null
					if (ID == null) {
						adapter.setID(eObj.toString());
						ret = eObj.toString();
					}
					else {
						adapter.setID(ID);
						ret = ID;
					}
				}
				return ret;
			}
		};
	}

	/**
	 * Returns the identifier of a woven element. The type of the element ref is passed as parameter, to be able to know which 
	 * identifier provider will be instantiated.<br/>
	 *  In the ATL transformation, the method is called in the following way:
	 *  <b>someElement.getElementIDbyRefType ('ElementRefXMI')</b>
	 * @param frame
	 * @param self
	 * @return
	 */
	private Operation getElementIDbyRefType() {
		return new Operation (2,"getElementIDbyRefType"){
			@Override
			public Object exec(AbstractStackFrame frame) {
				Object[] localVars = frame.getLocalVars();
				EObject eObj= null;
				String ret=null;
				String elementRef = null;
				if (localVars[0] instanceof EObject && localVars[1] instanceof String) {
					eObj = (EObject) localVars[0];
					elementRef = (String) localVars[1];
				}
				else
					return null;
				WeaverExtensionManager em = new WeaverExtensionManager(null);
				IIdentifierAdapter adapter = null;
				Object aclass = null;
				try {
					aclass = em.getProviderImplementationClassByType(elementRef);
				} catch (WeaverException e) {
					e.printStackTrace();
				}
				if (aclass instanceof IIdentifierAdapter)  {
					adapter = (IIdentifierAdapter) aclass;
					adapter.setTarget(eObj);
					ret = (String) adapter.getID();			
				} else {
					//if there is no adapter, gest the identifier directly
					ret =  eObj.eResource().getURIFragment(eObj);
				}
				return (ret == null) ? "" : ret;
			}
		};
	}


	/**
	 * Returns the woven model descriptor of a given woven model identification 
	 * @param refModel
	 * @param xml the xml file where the descriptor will be built
	 * @return
	 */
	private static WovenModelDescriptor getModelDescriptor(String refModel, WeaverXMLMetadata xml) {
		WovenModelDescriptor descriptor = null;
		for (Iterator wovens = xml.getWovenModels().iterator(); wovens.hasNext();) {
			String aPath = null;
			descriptor = (WovenModelDescriptor) wovens.next();

			if (descriptor.isModelExists())
				aPath = descriptor.getModelFile();
			else
				aPath = descriptor.getMetamodelFile();

			if (refModel.substring( refModel.indexOf("/") ).equals( aPath ) ) {				
				return descriptor;						
			}
		}
		return null;
	}	
	
	
	/**
	 * Returns a woven model, given the map of models, its ID and the ID in the current descriptor
	 * @param map
	 * @param modelURI
	 * @param modelIDNumber
	 * @return
	 */
	private static EMFModel getWovenASMModel (Iterator<IModel> inModels, String modelURI, String modelIDNumber  ){
		EMFModel woven = null;
		woven = null;

		for (Iterator models = inModels; models.hasNext();) {
			woven = (EMFModel) models.next();
			Resource auxRes = woven.getResource();
			if (auxRes != null) {
				ResourceIdentifierAdapter adapter = (ResourceIdentifierAdapter) WeaverUtil.getIdentifierAdapter(auxRes);
				if (adapter == null) {
					adapter = new ResourceIdentifierAdapter(auxRes);
				}
				if (adapter.getID().equals( modelURI ) ) {
					return woven;
				}
			}
		}	
		return null;
	}
	
	/**
	 * Returns the model element referred by an elementRef, and with a given ID 
	 * @param res
	 * @param elementRef
	 * @param refElementStr
	 * @param woven
	 * @return
	 */
	private static EObject getModelElement (Resource res, EObject elementRef, String refElementStr) {
		EObject ret = null;
		if (res != null) {
			WeaverExtensionManager em = new WeaverExtensionManager(null);
			IIdentifierAdapter adapter = null;
			try {
				Object aclass = em.getProviderImplementationClass(elementRef);
				if (aclass instanceof IIdentifierAdapter)  {
					adapter = (IIdentifierAdapter) aclass;
					for (Iterator eobjs = res.getAllContents(); eobjs.hasNext();) {
						Object element =  eobjs.next();
						if (element instanceof EObject) {
							adapter.setTarget((EObject)element);
							if ( adapter.getID().equals(refElementStr) ) {
								ret = (EObject) element;
								return ret;
							}
						} 
					}
				} else {
					//if there is no adapter, gest the element directly
					EObject eobj = res.getEObject(refElementStr);
					ret = eobj;
				}
					
			} catch (WeaverException e) {
				e.printStackTrace();
			}
		}		
		return ret;
	}	
	
	/**
	* Returns a similarity measure between two strings. Method used: Levenshtein distance. Requires the
	* simmetrics API (from UK Sheffield University). The jar file must be placed in the lib folder of the plug-in, and renamed
	* to simmetrics.jar. It is available at http://sourceforge.net/projects/simmetrics/.<br/>
	* In the ATL transformation, the method is called in the following way:
	* <b>'someString'.getStringSim ('anotherString')</b>
	* @param frame
	* @param left
	* @param right
	* @return
	*/
	
	private Operation getStringSim() {
		return new Operation (2,"getStringSim"){
			@Override
			public Object exec(AbstractStackFrame frame) {
				Object[] localVars = frame.getLocalVars();
				String leftStr = null;
				String rightStr = null;				
				if (localVars[0] instanceof String && localVars[1] instanceof String) {
					leftStr = (String) localVars[0];
					rightStr = (String) localVars[1];
				}
				else
					return null;
				Levenshtein lvh = new Levenshtein();
//				//Soundex s = new Soundex();
//				//QGramsDistance qd = new QGramsDistance();
				double realNumb = 0;
				if (!leftStr.equals("") && !rightStr.equals(""))
				realNumb = lvh.getSimilarity(leftStr, rightStr);
//				// + s.getSimilarity(leftStr, rightStr)
//				// qd.getSimilarity(leftStr,rightStr);
				return realNumb;
			}
		};
	}	
	/**
	 *  Generates the value of the "ref" attribute for a woven model. In other words, generates an identifier for models.
	 *  For the moment, models are identified uniquely by their URIs. <br/>
	 *  In the transformation, the method is called in the following way:
	 * 		<b>MOF!EPackage.generateModelRef('left')</b>
	 * 
	 * @param frame parameter set up internally by the ATL engine
	 * @param self
	 * @param wovenModel
	 * @return
	 */
	
	private Operation generateModelRef() {
		return new Operation (2,"generateModelRef"){
			@Override
			public Object exec(AbstractStackFrame frame) {
				Object[] localVars = frame.getLocalVars();
				EObject eObj= null, ret=null;
				String aString = null;
				if (localVars[0] instanceof EObject && localVars[1] instanceof String) {
					eObj = (EObject) localVars[0];
					aString = (String) localVars[1];
				}
				else
					return null;
				
				EMFModel model = (EMFModel) frame.getExecEnv().getModelsByName().get(aString);
								
				ResourceIdentifierAdapter resAdapter = (ResourceIdentifierAdapter) WeaverUtil.getIdentifierAdapter(model.getResource());
				if (resAdapter == null) {
					resAdapter = new ResourceIdentifierAdapter(model.getResource());
				}	
				return resAdapter.getID();
			}
		};
	}	
	/**
	 * This method takes as input a WLinkEnd of a weaving model and it returns the referred element of a woven model <br/>
	 * In the transformation, the method is called in the following way:
	 * 		<b>someLinkEnd.getReferredElement()</b>
	 * @param frame parameter set up internally by the ATL engine
	 * @param self the model element
	 * @return 
	 */
	//StackFrame frame, ASMEMFModelElement self
	private Operation getReferredElement() {
		return new Operation (1,"getReferredElement"){
			@Override
			public Object exec(AbstractStackFrame frame) {
				Object[] localVars = frame.getLocalVars();
				EObject eObj= null, ret=null;
				if (localVars[0] instanceof EObject) {
					eObj = (EObject) localVars[0];
				}
				else
					return null;
				WeaverXMLMetadata xml = getXmlFile(eObj);
				EObject linkEnd = eObj;
				Object elementRef = WeaverUtil.get(linkEnd, "element");
				if (elementRef instanceof EObject){
					EObject wModelRef = ((EObject) elementRef).eContainer();
					if (wModelRef != null){
						//reference of the object
						String refElementStr = (String) WeaverUtil.get((EObject) elementRef, "ref");
						String refModel = (String) WeaverUtil.get(wModelRef, "ref");						
						//ResourceIdentifierAdapter adapter = ;
						WovenModelDescriptor descriptor = getModelDescriptor(refModel, xml);
						if (descriptor != null) {

							EMFModel wovenM = getWovenASMModel(frame.getExecEnv().getModels(), refModel, descriptor.getId());
							Resource res = wovenM.getResource();					
							ret = getModelElement(res, (EObject) elementRef, refElementStr);
						}
					}
				}		
				return (ret == null) ? OclUndefined.SINGLETON : ret;
			}
		};
	}	
	
	/**
	 * Returns the identifier (ID) of a woven element. For the moment it is necessary to pass as parameter the related weaving model. This is the 
	 * way we find which WElementRef is used, and which provider class is called.<br/>
	 *  In the ATL transformation, the method is called in the following way:
	 *  	<b>someElementConformingToMof.getElementID('IN')</b>
	 * @param frame
	 * @param self
	 * @return
	 */
	
	private Operation getElementID() {
		return new Operation (2,"getElementID"){
			@Override
			public Object exec(AbstractStackFrame frame) {
				Object[] localVars = frame.getLocalVars();
				EObject eObj= null;
				String ret=null;
				String weavingModelName = null;
				if (localVars[0] instanceof EObject && localVars[1] instanceof String) {
					eObj = (EObject) localVars[0];
					weavingModelName = (String) localVars[1];
				}
				else
					return null;
				
				Resource wovenRes = eObj.eResource();
				
				EMFModel weaving = (EMFModel) frame.getExecEnv().getModelsByName().get(weavingModelName);
				
				ResourceIdentifierAdapter resAdapter = (ResourceIdentifierAdapter) WeaverUtil.getIdentifierAdapter(wovenRes);
				String modelURI = null;
				if (resAdapter == null) {
					resAdapter = new ResourceIdentifierAdapter(wovenRes);
					modelURI = (String) resAdapter.getID();
				}
				//searches for the WModelRef
				List wmodelRefs = WeaverUtil.getElementsByTypeHierarchy(WeaverConstant.MODEL_REF_ELEMENT, weaving.getResource());
				
				EObject wmodelRef = null;
				for (Iterator mRefs = wmodelRefs.iterator(); mRefs.hasNext();) {
					wmodelRef = (EObject) mRefs.next();
					String aUri = (String)wmodelRef.eGet( wmodelRef.eClass(). getEStructuralFeature("ref"));
					if ( aUri.equals(modelURI) )
						break;
				}
				//WeaverUtil.getAllMatchedEObjs(weaving, "WModelRef", featureName, featValue)
				String welementRefName = WeaverUtil.getAnnotationRefType(wmodelRef.eClass());
				WeaverExtensionManager em = new WeaverExtensionManager(null);
				IIdentifierAdapter adapter = null;
				Object aclass = null;
				try {
					aclass = em.getProviderImplementationClassByType(welementRefName);
				} catch (WeaverException e) {
					e.printStackTrace();
				}
				if (aclass instanceof IIdentifierAdapter)  {
					adapter = (IIdentifierAdapter) aclass;
					adapter.setTarget(eObj);
					ret = (String) adapter.getID();			
				} else {
					//if there is no adapter, gest the element directly
					ret = ((XMIResourceImpl)wovenRes).getID(eObj);
				}
				return (ret == null) ? "" : ret;
			}
		};
	}

	
	/**
	 * returns the XML metadata file for a given weaving model 
	 * @param model the weaving model
	 * @return
	 */
	private static WeaverXMLMetadata getXmlFile (EObject eObj){
		Resource weavingModel = eObj.eResource();
		
		String txtPath = ""; 
		//uris created by ATL do not have platform:resource before (check this)
		if (weavingModel.getURI().toString().indexOf(WeaverConstant.RESOURCE_STRING) != -1) 
			txtPath = weavingModel.getURI().toString().substring( WeaverConstant.RESOURCE_STRING.length());
		else
			txtPath = weavingModel.getURI().toString();

		IWorkspaceRoot root = ResourcesPlugin.getWorkspace().getRoot();
		File properties = new File(root.getLocation() + txtPath + "."+ WeaverConstant.PROPERTY_LAST_EXT);
		//load the properties files path
		WeaverXMLMetadata xml = null;
		try {
			xml = new WeaverXMLMetadata(properties);
			xml.loadXMLConfigFile();
		} catch (WeaverException e) {
			e.printStackTrace();
		} catch (ParserConfigurationException e) {
			e.printStackTrace();
		} catch (SAXException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
		return xml;
		
	}
	
	
}
