/*******************************************************************************
 * Copyright (c) 2006 INRIA 
 * 2010 UFPR - C3SL.

 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 *******************************************************************************/

package org.eclipse.gmt.weaver.transformation.util;

import java.io.File;
import java.net.URL;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.FileLocator;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.Platform;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.gmt.weaver.core.WeaverModelManager;
import org.eclipse.gmt.weaver.core.WeaverXMLMetadata;
import org.eclipse.gmt.weaver.core.data.WovenModelDescriptor;
import org.eclipse.gmt.weaver.exception.WeaverException;
import org.eclipse.gmt.weaver.plugin.WeaverConstant;
import org.eclipse.gmt.weaver.plugin.WeaverPlugin;
import org.eclipse.gmt.weaver.plugin.util.WeaverUtil;
import org.eclipse.gmt.weaver.transformation.TransformationActivator;
import org.eclipse.gmt.weaver.transformation.config.TBinding;
import org.eclipse.gmt.weaver.transformation.config.TransformationConfig;
import org.eclipse.jface.dialogs.InputDialog;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.m2m.atl.core.IExtractor;
import org.eclipse.m2m.atl.core.IModel;
import org.eclipse.m2m.atl.core.IReferenceModel;
import org.eclipse.m2m.atl.core.ModelFactory;
import org.eclipse.m2m.atl.core.emf.EMFInjector;
import org.eclipse.m2m.atl.core.emf.EMFModel;
import org.eclipse.m2m.atl.core.launch.ILauncher;
import org.eclipse.m2m.atl.core.service.CoreService;

public class TransformationUtil {

	/**
	 * executes a transformation (higher order or matching) available in the weaving graphical interface.
	 * @param models the map of models
	 * @param wManager the model manager
	 * @param fileXML the xml file associated with the weaving model that is being edited
	 * @param tc the configuration of the transformation that will be executed
	 * @return
	 * @throws WeaverException
	 * 
	 * @author - Marcos Didonet Del Fabro (LINA Labs, INRIA team), UFPR - C3SL<br>
	 * 	@since 1.0 
	 */
 public static IModel executeTransfo(ILauncher launcher, WeaverModelManager wManager,
		 WeaverXMLMetadata fileXML, TransformationConfig tc) throws WeaverException {
	 IModel ret = null;
	 try {    		
		 //charge the auxiliary structures 
		 ModelFactory factory = CoreService.getModelFactory(WeaverConstant.EMF);
		 EMFInjector injector = (EMFInjector) CoreService.getInjector(WeaverConstant.EMF);
		 
		 URL tPlatform = Platform.getBundle(tc.plugin).getEntry(tc.transformation);		 
		 URL transfoURL = null;
		 transfoURL = FileLocator.toFileURL(tPlatform);           
		 //now get the woven models
		 WeaverXMLMetadata inXML = fileXML;
		 for (Iterator bindings = tc.bindings.iterator(); bindings.hasNext(); ) {
			 TBinding tb = (TBinding) bindings.next();
			 //TODO change this code to navigate through the weaving model, not the metadata file		
			 if (! tb.isExtraBinding ){

				 List wovenModels = inXML.getWovenModels();										
				 Iterator aux = wManager.getWovenResources().iterator();

				 for (Iterator wModels = wovenModels.iterator(); wModels.hasNext();) {
					 WovenModelDescriptor woven = (WovenModelDescriptor) wModels.next();
					 Resource wRes = (Resource) aux.next();
					 if (woven.getReferenceName().equals(tb.weavingReference)) {
						 //checks which reference model will be loaded
						 IReferenceModel wovenMetamodel; 
						 //wovenEMF = ASMEMFModel.loadASMEMFModel(tb.header, (ASMEMFModel)modelHandler.getMof(), wRes, null);
						 if (tb.metamodelHeader.equals(WeaverConstant.MOF)) {
							 wovenMetamodel = factory.getMetametamodel();
						 }
						 else {
							 wovenMetamodel = factory.newReferenceModel();
							 injector.inject(wovenMetamodel, wRes);
						 }
						 IModel wovenEMF = null;
						 wovenEMF = factory.newModel(wovenMetamodel);
						 injector.inject(wovenEMF, wRes);
						 launcher.addInModel(wovenEMF, tb.header, tb.metamodelHeader);
					 }
				 }		
			 }
			 else {//handle extra bindings that were not specified in the standard transformation signature
				 InputDialog msgDialog = new InputDialog(null,WeaverPlugin.getPlugin().getString("_UI_extra_model"),
						 tb.description,"",null);             		
				 if (msgDialog.open() == org.eclipse.jface.window.Window.OK){
					 URI uri = URI.createPlatformResourceURI(msgDialog.getValue());
					 if (launcher.getModel(tb.metamodelHeader)!=null){
						 IModel extraWoven = factory.newModel((IReferenceModel) launcher.getModel(tb.metamodelHeader)); 
					     injector.inject(extraWoven, uri.toString());
					     launcher.addInModel(extraWoven, tb.header, tb.metamodelHeader);
					 }
				 }
			 }
		 }
		 //EMF specific VM
		 launcher.launch(ILauncher.RUN_MODE, null, Collections.EMPTY_MAP, transfoURL.openStream());
		 ret = launcher.getModel("OUT");
		 if (ret == null) {// error when executing the transformation - this code should be refactored
			 MessageDialog.openError( null, "Error", 
			 "The input or output models of the transformation could not be configured.\n " +
			 "Please check \n " +
			 "1) the metamodel extension used may be compatible with the selected operation\n " +
			 "2) the weaving model structuration may not be supported by the HOT");					
		 }
		 
	 } catch (Exception e) {
		 if (ret == null) {// error when executing the transformation - this code should be refactored
			 MessageDialog.openError( null, "Error",
					 "The input or output models of the transformation could not be configured.\n " +
					 "Please check\n " +
					 "1) the metamodel extension used may be compatible with the selected operation\n " +
					 "2) the weaving model structuration may not be supported by the HOT\n\n " +	
					 "Exception message: " + e.getMessage());					
		 }
		 throw new WeaverException(IStatus.ERROR, 
				 "The input or output models of the transformation could not be configured.\n " +
				 "Please check\n " +
				 "1) the metamodel extension used may be compatible with the selected operation\n " +
				 "2) the weaving model structuration may not be supported by the HOT\n\n " +	
				 "Exception message: " + e.getMessage());
	 }		
	 return ret;
 }	 
 /**
	 * Saves a new weaving model (passed by in). It creates a new metadata file. The metadata file
	 * is created based on the weaving model that is being edited (available in the model manager).
	 * The transformation config element contain naming settings   
	 * @param tc
	 * @param in
	 * @param wManager
	 * @return
	 */
	 public static IModel saveNewModel (TransformationConfig tc, IModel in, WeaverModelManager wManager, WeaverXMLMetadata currentXML,String extraSuffix) {
	    	IModel output = null;
	    	try {	    	
	    		IExtractor extractor = CoreService.getExtractor(WeaverConstant.EMF);
	    		Resource wModel = wManager.getWeavingModel();
	    		//AtlModelHandler modelHandler = AtlModelHandler.getDefault(TransformationConstant.AMWModelHandler);

	    		WeaverXMLMetadata inXML = currentXML;
	    		String inModelUri;
	    		if (tc.type.equals(TransformationConstant.HOT_TRANSFORMATION)) 
	    			inModelUri = wModel.getURI().trimFileExtension().path().replaceAll("resource/", "") + "_" + tc.suffix+extraSuffix+".ecore";
	    		else {
	    			inModelUri = wModel.getURI().trimFileExtension().path().replaceAll("resource/", "") + "_" + tc.suffix+extraSuffix+".amw";
	    			//generates the metadatafile
	    			IPath path = new Path(inModelUri); 
	    			IFile file = ResourcesPlugin.getWorkspace().getRoot().getFile(path);
	    			
	    			File propertiesFile = WeaverUtil.getPropertyFilePath(file).toFile();
	    			WeaverXMLMetadata xmlFileTemp = new WeaverXMLMetadata(propertiesFile);
	    			xmlFileTemp.createXMLConfigFile(
	    					inXML.getWovenModels(),
	    					inModelUri,
	    					inXML.getWeavingModelPanel(),
	    					inXML.getWModelName(),
	    					inXML.getMetamodelExtension(),
	    					inXML.getLocalMetamodelExtension(),
	    					inXML.getWeavingEcore() );				        			    			
	    		}
	    		extractor.extract(in, inModelUri);
	    		//if the output is an ATL model, it also extract the model into the ATL textual format
	    		if (tc.suffix.equals("hot") || tc.suffix.equals("atl")){
	    			IExtractor ATLExtractor = CoreService.getExtractor(WeaverConstant.ATL);
	    			
	    			ATLExtractor.extract(in, ResourcesPlugin.getWorkspace().getRoot().getLocation().toString()+ 
	    					wModel.getURI().trimFileExtension().path().replaceAll("resource/", "") +".atl");
	    		}
	    		output = in;
	    	} catch (Exception e) {
	    		e.printStackTrace();
	    	}		
	    	return output;
	    }
/**
 * this method is also used by the global weaver editor. It is not put in the Weaver Util because it is sets a very specific configuration
 * type
 * @param tc
 * @param inModel
 * @return
 */
	 public static ILauncher setModelsMap (IModel inModel, WeaverModelManager wManager, boolean isMatching, 
			 boolean hasMatchConfig, IModel cModel, IModel cMetamodel ) {
		 IModel in;
		 IReferenceModel amwMM;
		 IModel amwM;
		 ModelFactory factory;
		 EMFInjector injector;
		 ILauncher launcher = null;
		 try {
			 factory = CoreService.getModelFactory(WeaverConstant.EMF);
			 injector = (EMFInjector) CoreService.getInjector(WeaverConstant.EMF);
			 launcher = CoreService.getLauncher(WeaverConstant.EMF_VM);
			 launcher.initialize(Collections.<String, Object> emptyMap());
			 
			 			
			 if (inModel == null) { //this means it is executed a single time, not with the configuration model
				 amwMM = factory.newReferenceModel();
				 injector.inject(amwMM, wManager.getWeavingMetamodel());
				 in = factory.newModel(amwMM);
				 injector.inject(in, wManager.getWeavingModel());
			 }
			 else {				
				 amwMM = inModel.getReferenceModel();
				 in = inModel;
			 }		
			 launcher.addInModel(in, "IN", "AMW");

			 IModel out;
			 String referenceModelName;
			 if (isMatching){
				 out = factory.newModel(amwMM);
				 referenceModelName = "AMW";
			 } else {    		
				 referenceModelName = "ATL";
				 out = factory.newModel(factory.getBuiltInResource("ATL.ecore"));
			 }
			 launcher.addOutModel(out, "OUT", referenceModelName);

			 //set up the configuration model
			 if (hasMatchConfig) {
				 launcher.addInModel(cModel, "match", "match_p");
			 }
		 } catch (WeaverException e) {
			 e.printStackTrace();
		 } catch (Exception e) {
			 e.printStackTrace();
		 }
		 return launcher;		
	 }

}
