/*******************************************************************************
 * Copyright (c) 2005 INRIA.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Marcos Didonet Del Fabro - initial API and implementation
 *    Guillaume Gueltas
 *******************************************************************************/


package org.eclipse.gmt.weaver.core;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IExtension;
import org.eclipse.core.runtime.IExtensionPoint;
import org.eclipse.core.runtime.IExtensionRegistry;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Platform;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.impl.EClassImpl;
import org.eclipse.gmt.weaver.core.data.WeaverExtensionPointDescriptor;
import org.eclipse.gmt.weaver.editors.interfaces.IWeaverItemProvider;
import org.eclipse.gmt.weaver.exception.WeaverException;
import org.eclipse.gmt.weaver.plugin.WeaverConstant;
import org.eclipse.gmt.weaver.plugin.WeaverPlugin;
import org.eclipse.gmt.weaver.plugin.util.WeaverUtil;

/**
 * This class is a manager of extensions for the weaver
 * @since 1.0
 */
public class WeaverExtensionManager {
   
	protected WeaverModelManager modelManager;
	// list of "extension points descriptors" in the application
	// each "extension points descriptor" have a list of extensions which use this extension point
    protected List extensionPoints = new ArrayList();
    
    // Map of icons
    private Map iconMap = new HashMap();
   
    /**
     * Create a new WeaverExtensions
     */
    public WeaverExtensionManager (WeaverModelManager manager) {
    	modelManager = manager;
        extensionPoints.add(new WeaverExtensionPointDescriptor(WeaverConstant.METAMODEL_EXT, WeaverConstant.METAMODEL_EXT_ID));
        extensionPoints.add(new WeaverExtensionPointDescriptor(WeaverConstant.ITEM_PROVIDER_EXT, WeaverConstant.ITEM_PROVIDER_EXT_ID));
        extensionPoints.add(new WeaverExtensionPointDescriptor(WeaverConstant.WOVEN_PANEL_EXT, WeaverConstant.WOVEN_PANEL_EXT_ID));
        extensionPoints.add(new WeaverExtensionPointDescriptor(WeaverConstant.WEAVING_PANEL_EXT, WeaverConstant.WEAVING_PANEL_EXT_ID));
        extensionPoints.add(new WeaverExtensionPointDescriptor(WeaverConstant.INITIALIZE_EXT, WeaverConstant.INITIALIZE_EXT_ID));        
    }
    /**
     * Adds a new extension into the selected extensions list from this extension point descriptor
     * An extension point has N extensions that use it. This object is used to manage a list of some 
     * selected extensions.
     * 
     * @param EXTENSION_POINT_ID  predefined constant indicating the extension point ID
     * @param extension the extension
     */
    public void addExtensionToDescriptor(int EXTENSION_POINT_LOCAL_ID, IExtension extension){
        WeaverExtensionPointDescriptor extenstionPoint = (WeaverExtensionPointDescriptor)extensionPoints.get(EXTENSION_POINT_LOCAL_ID);           
        if (extenstionPoint!=null)
        	extenstionPoint.addSelectedExtension(extension);            
    }
    /**
     * Removes a extension from this extension point descriptor
     * @param EXTENSION_POINT_ID predefined constant indicating the extension point ID
     * @param extension the extension
     */
    public void removeExtensionFromDescriptor(int EXTENSION_POINT_LOCAL_ID, IExtension extension){
        WeaverExtensionPointDescriptor extenstionPoint = (WeaverExtensionPointDescriptor)extensionPoints.get(EXTENSION_POINT_LOCAL_ID);
        if (extenstionPoint!=null)
        	extenstionPoint.removeSelectedExtension(extension);
    }

    /**
     * Adds all extensions 
     * @param extDescriptor
     */
    protected void addAllExtensionsInDescriptor(WeaverExtensionPointDescriptor extDescriptor){
		IExtension[] ext = getAllExtensions(extDescriptor.getExtensionPointID());
		for (int i = 0; i < ext.length; i++){
		    extDescriptor.addSelectedExtension(ext[i]);
		}	        
    }
    
    /**
     * Returns a List of file Path with all km3 files declared in the extension
     * the file path is in a array : {extension path, file}
     * @return
     * @throws WeaverException 
     */
    public Collection getPluginMetamodelPath(){
		Collection paths = new ArrayList();		
		addAllExtensionsInDescriptor((WeaverExtensionPointDescriptor)extensionPoints.get(WeaverConstant.METAMODEL_EXT));        
		WeaverExtensionPointDescriptor extDescriptor =  (WeaverExtensionPointDescriptor)extensionPoints.get(WeaverConstant.METAMODEL_EXT);
		for (Iterator i = extDescriptor.getSelectedExtensions().iterator(); i.hasNext(); ){		    
			IExtension extension = (IExtension)i.next();
			
			//TODO verify the path
			for(Iterator iter=getAllConfigurationElements(extension,"extensionFile").iterator(); iter.hasNext();){ 
				String[] array = new String[2];
				array[0] = extension.getNamespaceIdentifier();
				IConfigurationElement confElement = (IConfigurationElement) iter.next();           			            
				String extensionName = confElement.getAttribute("relativePath")+confElement.getAttribute("fileName");			        
				array[1] = extensionName.replaceAll("update@/","");
				array[1] = array[1].replaceAll("update@","");;
				paths.add(array);
			}
		}
		return paths;	
	}
    /**
     * Gets a  collection of configuration element from an extension based on its name
     * @param extension
     * @param name
     * @return
     */
    private Collection getAllConfigurationElements(IExtension extension, String name){
    	Collection list = new ArrayList();
        IConfigurationElement[] elements = extension.getConfigurationElements();
        for (int j=0; j < elements.length; j++){
            if (elements[j].getName().equals(name))
            	list.add(elements[j]);
        }
        return list;        
    }
    
    /**
     * Gets the first configuration element from an extension based on its name
     * @param extension
     * @param name
     * @return
     */
    private IConfigurationElement getFirstConfigurationElement(IExtension extension, String name){
        IConfigurationElement[] elements = extension.getConfigurationElements();
        for (int j=0; j < elements.length; j++){
            if (elements[j].getName().equals(name))
            	return elements[j];
        }
        return null;        
    }   
    
    /**
     * Loads list of extensions in the associated extension point descriptor
     * @param the list of extension selected (String)
     * @param extensionPointID the identifier of the extension point id
     */
    public void loadExtensions(List extensions, String extensionPointID){
        IExtension[] allLoadedExtensions = getAllExtensions(extensionPointID);
        // if our (all) String extension is in the list, we add it in the descriptor
        for (int i = 0; i < allLoadedExtensions.length; i++){
            if (extensions.contains(allLoadedExtensions[i].getSimpleIdentifier()))
                addExtensionToDescriptor(getLocalIdentifier(extensionPointID),allLoadedExtensions[i]);                    
        }
    }
    /**
     * Loads a extension in the associated extension point descriptor
     * @param the list of extension selected (String)
     * @param extensionPointID the identifier of the extension point id
     */
    public void loadExtension(String extension, String extensionPointID){
        IExtension[] allLoadedExtensions = getAllExtensions(extensionPointID);
        for (int i = 0; i < allLoadedExtensions.length; i++){
            if (extension.equals(allLoadedExtensions[i].getSimpleIdentifier()))
                addExtensionToDescriptor(getLocalIdentifier(extensionPointID),allLoadedExtensions[i]);                    
        }
    }
    
    /**
     * Gets the local ID associated with the extension description
     * @param extensionPointID
     * @return
     */
    private int getLocalIdentifier(String extensionPointID){
        for (Iterator ext = extensionPoints.iterator(); ext.hasNext();){
            WeaverExtensionPointDescriptor descriptor = (WeaverExtensionPointDescriptor) ext.next();
            if (descriptor.getExtensionPointID() == extensionPointID)
                return descriptor.getExtensionLocalIdentifier();         
        }
        return 0;
    }    
    
    /**
     * Returns all the extensions based on this id
     * that the registry has found
     * @param extensionPointID
     * @return 
     */
    public static IExtension[] getAllExtensions(String extensionPointID){
		IExtensionRegistry registry = Platform.getExtensionRegistry();
		IExtensionPoint point = registry.getExtensionPoint(extensionPointID);
		IExtension[] extensions = point.getExtensions();
		return extensions;
    }   
    
    /**
     * Gets the first item provider adapter declared in the extensions points for the type. It does not search in the parent classes 
     * @param className
     * @return
     * @throws WeaverException
     */
    public Object getProviderImplementationClassByType(String className) throws WeaverException{
    	addAllExtensionsInDescriptor((WeaverExtensionPointDescriptor)extensionPoints.get(WeaverConstant.ITEM_PROVIDER_EXT));        
    	WeaverExtensionPointDescriptor extDescriptor = (WeaverExtensionPointDescriptor)extensionPoints.get(WeaverConstant.ITEM_PROVIDER_EXT);
    	for (Iterator i = extDescriptor.getSelectedExtensions().iterator(); i.hasNext(); ){
    		IExtension ext = (IExtension)i.next();
    		IConfigurationElement confElement = getFirstConfigurationElement(ext,"itemProviderAdapter");            
    		if (confElement !=null){
    			String adaptedClassName = "";
    			// verifies if it is the default item provider to the given object
    			if (confElement.getAttribute("default").equals("true")){
    				adaptedClassName = confElement.getAttribute("adaptedClassName");            		
    				if (adaptedClassName.equals(className)){ // found the class
    					try {
    						Object plugin = confElement. createExecutableExtension("class");
    						if (plugin != null){
    							IWeaverItemProvider provider = (IWeaverItemProvider)plugin;
    							provider.setIconMap(iconMap);
    							return provider;
    						}
    						return plugin;
    					} catch (CoreException e){
    						throw new WeaverException(IStatus.ERROR,WeaverPlugin.getPlugin().getString("_UI_wizard_error_extension_2"),e);
    					}                                                    
    				}
    				//TODO does not check for children providers in this case : verify				
    			}
    		}            
    	}
    	return null;
    }    
    
    private String getAdaptedClassName(IExtension ext){
		IConfigurationElement confElement = getFirstConfigurationElement(ext,"itemProviderAdapter");            
		String adaptedClassName = "";
		if (confElement != null){
			if (confElement.getAttribute("default").equals("true") )
				adaptedClassName = confElement.getAttribute("adaptedClassName");
		}
		return adaptedClassName;    	
    }
    
    /**
     * returns the configuration element of the provider implementation class for a given object
     * @param extDescriptor
     * @param eObj
     * @return
     * @throws WeaverException
     */
    
    private IConfigurationElement findProviderConfElement (WeaverExtensionPointDescriptor extDescriptor, EClass eClass) throws WeaverException {
    	IConfigurationElement retConfElement = null;
    	String className = null; 
    	
    	//search for a class with the same name
    	for (Iterator i = extDescriptor.getSelectedExtensions().iterator(); i.hasNext(); ) {    		
    		IExtension ext = (IExtension)i.next();
    		//this extension is supposed to have only one configuration element
    		className = getAdaptedClassName(ext);
    		if (className.equals(eClass.getName())){
    			retConfElement = getFirstConfigurationElement(ext,"itemProviderAdapter");
   				break;
    		}	
    	}
    	 //searches in the super classes
    	if (retConfElement == null) {
    		for (Iterator i = eClass.getEAllSuperTypes().iterator(); i.hasNext(); ){
    			EClass aClass = (EClass)i.next();
    	    	for (Iterator iExt = extDescriptor.getSelectedExtensions().iterator(); iExt.hasNext(); ) {    		
    	    		IExtension ext = (IExtension)iExt.next();
    	    		//this extension is supposed to have only one configuration element
    	    		className = getAdaptedClassName(ext);
    	    		if (className.equals(aClass.getName())){
    	    			IConfigurationElement auxConf = getFirstConfigurationElement(ext,"itemProviderAdapter");
    	    			if (auxConf.getAttribute("isChildrenProvider").equals("true")){
    	    				retConfElement = auxConf;
    	    				break;
    	    			}
    	    		}
    	    	}
    	    	if (retConfElement != null)
    	    		break;
    		}
    	}  
    	return retConfElement;
    }
    
    /**
     * sets ups the icon map for a given class and configuration element
     * @param retConfElement
     * @param className
     * @throws WeaverException 
     */
    private void setProviderIcon(IConfigurationElement retConfElement, String className) throws WeaverException{
    	if (retConfElement != null && modelManager!=null) {
    		EClassifier classif = WeaverUtil.getClassifier(className,modelManager.getWeavingMetamodel());
			if (classif != null){//set icons for a class if it is contained in the metamodel
				String icon = retConfElement.getAttribute("icon");
				if(icon != null){
					iconMap.put(className, icon);
				}
			}
    	}
    }
    
    /**
     * Gets the first item provider adapter declared in the extensions points for this class
     * @param adaptedClassName
     * @return
     * @throws WeaverException
     */
    public Object getProviderImplementationClass(EObject eObj) throws WeaverException{
    	addAllExtensionsInDescriptor((WeaverExtensionPointDescriptor)extensionPoints.get(WeaverConstant.ITEM_PROVIDER_EXT));        
    	WeaverExtensionPointDescriptor extDescriptor = (WeaverExtensionPointDescriptor)extensionPoints.get(WeaverConstant.ITEM_PROVIDER_EXT);
    	//
    	IConfigurationElement confElement = findProviderConfElement(extDescriptor, eObj.eClass());
    	if (confElement != null)
			try {//instantiates the class
				setProviderIcon(confElement, eObj.eClass().getName());
				Object plugin = confElement.createExecutableExtension("class");
				if (plugin != null){
					IWeaverItemProvider provider = (IWeaverItemProvider)plugin;
					provider.setIconMap(iconMap);
					return provider;
				}
				return plugin;
			} catch (CoreException e){
				throw new WeaverException(IStatus.ERROR,WeaverPlugin.getPlugin().getString("_UI_wizard_error_extension_2"),e);
			}                                                    
    	return null;
    }   
    /**
     * sets ups the icons for all metamodel elements
     * @throws WeaverException
     */
    public void getAllProviderImplementationIcon() throws WeaverException {
    	addAllExtensionsInDescriptor((WeaverExtensionPointDescriptor)extensionPoints.get(WeaverConstant.ITEM_PROVIDER_EXT));        
    	WeaverExtensionPointDescriptor extDescriptor = (WeaverExtensionPointDescriptor)extensionPoints.get(WeaverConstant.ITEM_PROVIDER_EXT);
    	//
    	for (Iterator it = modelManager.getWeavingMetamodel().getAllContents(); it.hasNext();) {
			EObject element = (EObject) it.next();
			if (element instanceof EClassImpl) {
				if (iconMap.get(((EClassImpl)element).getName()) == null){
					IConfigurationElement confElement = findProviderConfElement(extDescriptor,(EClassImpl)element);
					setProviderIcon(confElement, ((EClassImpl)element).getName());
				}
			}			
		}    	
    }
    
    /**
     * Returns the classes extended and used for initializing a new model
     * @return
     * @throws WeaverException
     */
    public List getInitializingClasses() throws WeaverException {
    	List initializingClasses = new ArrayList();
    	
    	addAllExtensionsInDescriptor((WeaverExtensionPointDescriptor)extensionPoints.get(WeaverConstant.INITIALIZE_EXT));        
        WeaverExtensionPointDescriptor extDescriptor = (WeaverExtensionPointDescriptor)extensionPoints.get(WeaverConstant.INITIALIZE_EXT);
        for (Iterator i = extDescriptor.getSelectedExtensions().iterator(); i.hasNext(); ){
            IExtension ext = (IExtension)i.next();
            IConfigurationElement confElement = getFirstConfigurationElement(ext,"extensionInitialization");
            if (confElement !=null){
                try {
                    Object plugin = confElement.createExecutableExtension("class");
                    initializingClasses.add(plugin);
                } catch (CoreException e){
                	throw new WeaverException(IStatus.ERROR,WeaverPlugin.getPlugin().getString("_UI_wizard_error_extension_3"),e);
                }                                                    
            }            
        }
        return initializingClasses;
    }
    
    /**
     * Gets the woven panel declared in the extensions points for this class
     * @return
     * @throws WeaverException
     */
    public Object getWovenPanelImplementationClass(String id) throws WeaverException{
        
        WeaverExtensionPointDescriptor extDescriptor = (WeaverExtensionPointDescriptor)extensionPoints.get(WeaverConstant.WOVEN_PANEL_EXT);
        for (Iterator i = extDescriptor.getSelectedExtensions().iterator(); i.hasNext(); ){
            IExtension ext = (IExtension)i.next();
            
            if(ext.getSimpleIdentifier().equals(id)){
	            IConfigurationElement confElement = getFirstConfigurationElement(ext,"modelPanel");
	            if (confElement !=null){
	                try {
	                    Object plugin = confElement.createExecutableExtension("class");
	                    return plugin;
	                } catch (CoreException e){
	                	throw new WeaverException(IStatus.ERROR,WeaverPlugin.getPlugin().getString("_UI_wizard_error_extension_5"),e);
	                }                                                    
	            }   
            }
        }
        throw new WeaverException(IStatus.ERROR,WeaverPlugin.getPlugin().getString("_UI_wizard_error_extension_5"));
    }
    
    /**
     * Gets the weaving panel declared in the extensions points for this class
     * @return
     * @throws WeaverException
     */
    public Object getWeavingPanelImplementationClass(String id) throws WeaverException{
        
    	//addAllExtensionsInDescriptor((ExtensionPointDescriptor)extensionPoints.get(WeaverConstant.WEAVING_PANEL_EXT));        
        WeaverExtensionPointDescriptor extDescriptor = (WeaverExtensionPointDescriptor)extensionPoints.get(WeaverConstant.WEAVING_PANEL_EXT);
        for (Iterator i = extDescriptor.getSelectedExtensions().iterator(); i.hasNext(); ){
            IExtension ext = (IExtension)i.next();
            
            if(ext.getSimpleIdentifier().equals(id)){
	            IConfigurationElement confElement = getFirstConfigurationElement(ext,"weavingPanel");
	            if (confElement !=null){
	                try {
	                    Object plugin = confElement.createExecutableExtension("class");
	                    return plugin;
	                } catch (CoreException e){
	                	throw new WeaverException(IStatus.ERROR,WeaverPlugin.getPlugin().getString("_UI_wizard_error_extension_4"),e);
	                }                                                    
	            }   
            }
        }
        throw new WeaverException(IStatus.ERROR,WeaverPlugin.getPlugin().getString("_UI_wizard_error_extension_2"));
    }
}
