/*******************************************************************************
 * Copyright (c) 2005 INRIA.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Marcos Didonet Del Fabro - initial API and implementation
 *    Guillaume Gueltas
 *******************************************************************************/


package org.eclipse.gmt.weaver.core;


import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.eclipse.core.runtime.IStatus;
import org.eclipse.emf.ecore.EAnnotation;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.edit.domain.EditingDomain;
import org.eclipse.gmt.weaver.exception.WeaverException;
import org.eclipse.gmt.weaver.plugin.WeaverPlugin;
import org.eclipse.gmt.weaver.plugin.util.WeaverUtil;

/**
 * This Class contains help methods for manipulate EMF elements
 * @since 1.0
 */
public class WeaverModelHelper {
	
	private WeaverModelManager manager;
	protected int objectIndex;
	
	public WeaverModelHelper(WeaverModelManager manager){
		this.manager = manager;
		objectIndex = 0;
	}

	/**
	 * Creates a new EObject based in the weaving meta model (it does not add it into the main resource)
	 * @param className
	 * @param editingDomain
	 * @return
	 * @throws WeaverException 
	 * @throws WeaverException
	 */
	public EObject createWeavingElement(String className, EditingDomain editingDomain) throws WeaverException{
		Resource res = null;
		try{
			res = this.getManager().getWeavingMetamodel();
		}
		catch (Exception e) {}
		EObject instance = null;
		EClassifier classifier = WeaverUtil.getClassifier(className, res);		
		if (classifier instanceof EClass) 
			if (! ((EClass)classifier).isAbstract()){
				instance = classifier.getEPackage().getEFactoryInstance().create((EClass)classifier);
			}				
			else 
				throw new WeaverException(IStatus.ERROR,WeaverPlugin.getPlugin().getString("_UI_create_weaving_element_error"));
		return instance;
	}
	/**
	 * Creates a new EObject based in the weaving meta model and adds it into the resource)
	 * @param className
	 * @param editingDomain
	 * @return
	 * @throws WeaverException
	 */
	//TODO replace by createweavingElement after more testings. (changed for compability reasons with Eclipse 3.2
	public EObject createWeavingElement_32(String className, EditingDomain editingDomain){
		EObject instance = null;
		try {
			instance = createWeavingElement(className,editingDomain);
		} catch (WeaverException e) {}
		return instance;	
	}	
	
	/**
	 * Creates a new EObject based in the weaving meta model and adds it into the resource)
	 * @param className
	 * @param editingDomain
	 * @return
	 * @throws WeaverException
	 */
	public EObject createWeavingElementInRes(String className, EditingDomain editingDomain){
		EObject instance = null;
		try {
			instance = createWeavingElement(className,editingDomain);
			this.getManager().getWeavingModel(editingDomain).getContents().add(instance);			
		} catch (WeaverException e) {}
		return instance;	
	}
	
	/**
	 * Removes a weaving element
	 * @param eo
	 * @param editingDomain
	 */
	public void removeWeavingElement(EObject eo, EditingDomain editingDomain) {
		try {					
			this.getManager().getWeavingModel(editingDomain).getContents().remove(eo);
		} catch (WeaverException e) {}
	}	  
    
	/**
	 * Returns if this object is an instance of the className 
	 * @return true or false
	 */
	public boolean isInstanceOfMetaModel(EObject eObj, String className){
	    try {
			return WeaverUtil.isInstanceOf(eObj,className,this.getManager().getWeavingMetamodel());
		} catch (WeaverException e) {
			return false;
		}
	}      
	/**
	 * returns a list of derived  values
	 * @param eObj
	 * @param featureName
	 * @return
	 * @throws WeaverException
	 */
	public List getDerivedValues(EObject eObj, String featureName) throws WeaverException{
		List derivedList = new ArrayList();
		if (eObj != null){
			EStructuralFeature derivedFeat = eObj.eClass().getEStructuralFeature(featureName);
			if (derivedFeat != null){
				if (derivedFeat.isDerived()){
					String[] derivedStr = ((EAnnotation)derivedFeat.getEAnnotations().get(0)).getSource().split(",");
					EClass eClassObj =  eObj.eClass();
					List refList = eClassObj.getEReferences();
					for (int i = 0; i < derivedStr.length; i++){
						for (Iterator refs = refList.iterator(); refs.hasNext();){
							EReference subsetOf = (EReference) refs.next();
							if (subsetOf.getName().equals(derivedStr[i])){
								derivedList.addAll( (List)eObj.eGet(subsetOf) );
							}
						}
					}
				}
				else
					throw new WeaverException(IStatus.ERROR,WeaverPlugin.getPlugin().getString("_UI_manager_derived_error",derivedFeat.getName()));	
			}			
		}		
		return derivedList;		
	}
	/**
	 * @return Returns the manager.
	 */
	public WeaverModelManager getManager() {
		return manager;
	}
	/**
	 * @param manager The manager to set.
	 */
	public void setManager(WeaverModelManager manager) {
		this.manager = manager;
	}
}


