/*******************************************************************************
 * Copyright (c) 2005 INRIA.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Marcos Didonet Del Fabro - initial API and implementation
 *    Guillaume Gueltas
 *******************************************************************************/


package org.eclipse.gmt.weaver.editors;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.IPath;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.edit.domain.EditingDomain;
import org.eclipse.emf.edit.ui.action.EditingDomainActionBarContributor;
import org.eclipse.gmt.weaver.editors.actions.WeaverChildAction;
import org.eclipse.gmt.weaver.editors.actions.WeaverSiblingAction;
import org.eclipse.gmt.weaver.editors.interfaces.IWeaverMenuAdapter;
import org.eclipse.gmt.weaver.exception.WeaverException;
import org.eclipse.gmt.weaver.plugin.WeaverPlugin;
import org.eclipse.gmt.weaver.plugin.util.WeaverUtil;
import org.eclipse.jface.action.Action;
import org.eclipse.jface.action.ActionContributionItem;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.action.IContributionItem;
import org.eclipse.jface.action.IContributionManager;
import org.eclipse.jface.action.IMenuListener;
import org.eclipse.jface.action.IMenuManager;
import org.eclipse.jface.action.MenuManager;
import org.eclipse.jface.action.Separator;
import org.eclipse.jface.action.SubContributionItem;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.ISelectionChangedListener;
import org.eclipse.jface.viewers.ISelectionProvider;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.SelectionChangedEvent;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.dialogs.SaveAsDialog;

/**
 * Manages the installation/deinstallation of global actions for multi-page
 * editors. Responsible for the redirection of global actions to the active
 * editor. Multi-page contributor replaces the contributors for the individual
 * editors in the multi-page editor.
 * @since 1.0
 * 
 */
public class GlobalWeaverEditorContributor extends
        EditingDomainActionBarContributor implements ISelectionChangedListener {

    protected IEditorPart activeEditorPart;

    protected ISelectionProvider selectionProvider;

    // This will contain one {@link
    // org.eclipse.emf.edit.ui.action.CreateChildAction} corresponding to each
    // descriptor
    protected Collection createChildActions;

    // This is the menu manager into which menu contribution items should be
    // added for CreateChild actions.
    protected IMenuManager createChildMenuManager;

    // This will contain one {@link
    // org.eclipse.emf.edit.ui.action.CreateSiblingAction} corresponding to each
    // descriptor
    protected Collection createSiblingActions;

    //This is the menu manager into which menu contribution items should be
    // added for CreateSibling actions.
    protected IMenuManager createSiblingMenuManager;

    //actions for creating new EObjects    
    protected IWeaverMenuAdapter wMenuAdapter;

    /**
     * Action for display the property view
     */
    protected IAction showPropertiesViewAction = new Action(
    		WeaverPlugin.getPlugin().getString("_UI_Weaver_element_properties")) {
        public void run() {
            try {                
                getPage().showView("org.eclipse.ui.views.PropertySheet");
                ((GlobalWeaverEditor) activeEditor).getCreatePropertySheetPage().selectionChanged(activeEditor, 
                		selectionProvider.getSelection());            
                } catch (PartInitException exception) {
            }
        }
    };
    
    /**
     * Action to set the parent name equals to the selected child
     */
    protected IAction setParentNameAction = new Action(
    		WeaverPlugin.getPlugin().getString("_UI_Weaver_set_parent_name")) {
        public void run() {
            	ISelection selection = selectionProvider.getSelection();
            	if (selection instanceof IStructuredSelection){
            		Object obj = ((IStructuredSelection )selection).getFirstElement();
            		if (obj instanceof EObject){            			            			
            			EObject eParent = ((EObject)obj).eContainer();
            			//parent has name property            			
            			if (! (eParent.eClass().getEStructuralFeature("name") == null)){
            				WeaverUtil.setPropertyValue(eParent,
            						((EObject)obj).eGet(((EObject)obj).eClass().getEStructuralFeature("name")),
            						"name",((GlobalWeaverEditor) activeEditor).getEditingDomain());            				
            			}
            		}
            	}
        }
    };
    
    
    
	/**
     * Action to save the weaving metamodel as ecore 
     */
    protected IAction saveKm3MetamodelAction = new Action(
    		WeaverPlugin.getPlugin().getString("_UI_Save_weaving_metamodel_km3")) {
    	
    	public void run() {

    		GlobalWeaverEditor editorInstance = ((GlobalWeaverEditor)activeEditorPart);
    		SaveAsDialog saveAsDialog= new SaveAsDialog(editorInstance.getEditorSite().getShell());
    		saveAsDialog.open();
    		IPath path= saveAsDialog.getResult();
    		if (path != null) {
    			// create file
    			final IFile file = ResourcesPlugin.getWorkspace().getRoot().getFile(path);
    			if (file != null) {
    				try {
    					editorInstance.getModelManager().saveKm3WeavingMetamodel(file, editorInstance.getEditorSite().getShell());
    					
    				} catch (WeaverException e) {
    					e.printStackTrace();
    					MessageDialog.openWarning(editorInstance.getEditorSite().getShell(),
    							WeaverPlugin.getPlugin().getString("_UI_error_label"),
								e.getMessage());
    				}
    			}
    		}
    	}
    	
    };
    

	/**
     * Action to save the weaving metamodel as ecore 
     */
    protected IAction saveEcoreMetamodelAction = new Action(
    		WeaverPlugin.getPlugin().getString("_UI_Save_weaving_metamodel_ecore")) {
    	
        public void run() {
        	
        	GlobalWeaverEditor editorInstance = ((GlobalWeaverEditor)activeEditorPart);
			SaveAsDialog saveAsDialog= new SaveAsDialog(editorInstance.getEditorSite().getShell());
			saveAsDialog.open();
			IPath path= saveAsDialog.getResult();
			if (path != null) {
				// create file
				final IFile file = ResourcesPlugin.getWorkspace().getRoot().getFile(path);
				if (file != null) {
					try {
						editorInstance.getModelManager().saveEcoreWeavingMetamodel(file,editorInstance.getEditorSite().getShell());
					} catch (WeaverException e) {
						e.printStackTrace();
				    	MessageDialog.openWarning(editorInstance.getEditorSite().getShell(),
				    			WeaverPlugin.getPlugin().getString("_UI_error_label"),
				    			e.getMessage());
					}
				}
			}						
        }    	        
    };    
    
    /**
     * This creates an instance the contributor.
     */
    public GlobalWeaverEditorContributor() {    	
    }

    /* (non-Javadoc)
     * @see org.eclipse.ui.IEditorActionBarContributor#setActiveEditor(org.eclipse.ui.IEditorPart)
     */
    public void setActiveEditor(IEditorPart part) {
        super.setActiveEditor(part);      
        activeEditorPart = part;

        if (selectionProvider != null) {
            selectionProvider.removeSelectionChangedListener(this);
        }
        if (part == null) {
            selectionProvider = null;
        } else {
            selectionProvider = ((GlobalWeaverEditor) part).getWeaverEditorPanel().getViewer();
        	selectionProvider.addSelectionChangedListener(this);
            // Fake a selection changed event to update the menus.
            if (selectionProvider.getSelection() != null) {
                selectionChanged(new SelectionChangedEvent(selectionProvider,
                        selectionProvider.getSelection()));
            }
        }
    }
    
    /* (non-Javadoc)
     * @see org.eclipse.ui.part.EditorActionBarContributor#contributeToMenu(org.eclipse.jface.action.IMenuManager)
     */
    public void contributeToMenu(IMenuManager menuManager) {
        super.contributeToMenu(menuManager);
        IMenuManager submenuManager = new MenuManager(WeaverPlugin.getPlugin().getString(
                "_UI_Weaver_menu"), "weaver-main-menu");
        menuManager.insertAfter("additions", submenuManager);
        submenuManager.add(new Separator("settings"));
        submenuManager.add(new Separator("actions"));
        submenuManager.add(new Separator("additions"));
        submenuManager.add(new Separator("additions-end"));

        // Prepare for CreateChild item addition or removal.
        createChildMenuManager = new MenuManager(WeaverPlugin.getPlugin()
                .getString("_UI_New_child"));
        submenuManager.insertBefore("additions", createChildMenuManager);

        // Prepare for CreateSibling item addition or removal.
        createSiblingMenuManager = new MenuManager(WeaverPlugin.getPlugin()
                .getString("_UI_New_sibling"));
        submenuManager.insertBefore("additions", createSiblingMenuManager);
        
        // Force an update because Eclipse hides empty menus now.
        submenuManager.addMenuListener(new IMenuListener() {
            public void menuAboutToShow(IMenuManager menuManager) {
                menuManager.updateAll(true);
            }
        }); 
					
        addGlobalActions(submenuManager);
    }
    

    /* (non-Javadoc)
     * @see org.eclipse.jface.viewers.ISelectionChangedListener#selectionChanged(org.eclipse.jface.viewers.SelectionChangedEvent)
     */
    public void selectionChanged(SelectionChangedEvent event) {
        // Remove any menu items for old selection.
        if (createChildMenuManager != null) {
            depopulateManager(createChildMenuManager, createChildActions);
        }
        if (createSiblingMenuManager != null) {
            depopulateManager(createSiblingMenuManager, createSiblingActions);
        }
        // Query the new selection for appropriate new child/sibling descriptors
        Collection newChildDescriptors = null;
        Collection newSiblingDescriptors = null;
        Object object = null;
        
        ISelection selection = event.getSelection();
        if (selection instanceof IStructuredSelection
                && ((IStructuredSelection) selection).size() == 1) {
            object = ((IStructuredSelection) selection).getFirstElement();

            EditingDomain domain = ((GlobalWeaverEditor) activeEditor)
                    .getEditingDomain();

           newChildDescriptors = domain.getNewChildDescriptors(object, null);
           newSiblingDescriptors = domain.getNewChildDescriptors(null, object);           
        }
        // Generate actions for selection; populate and redraw the menus.
        //TODO these actions should be centralized in the WeavingPanel class
        createChildActions = generateCreateChildActions(newChildDescriptors,
                selection);
        createSiblingActions = generateCreateSiblingActions(
                newSiblingDescriptors, selection);

        if (createChildMenuManager != null) {
            populateManager(createChildMenuManager, createChildActions,
             null);
            createChildMenuManager.update(true);
        }
        if (createSiblingMenuManager != null) {
            populateManager(createSiblingMenuManager, createSiblingActions,
             null);
            createSiblingMenuManager.update(true);
        }
        // Prepare WLinkEnd actions to create a new WElementRef
        if (object instanceof EObject) {
             if (wMenuAdapter == null)
                 wMenuAdapter = ((GlobalWeaverEditor) activeEditorPart).getWeaverEditorPanel();
            wMenuAdapter.createActions((EObject) object);
        }
    }

    /**
     * This removes from the specified manager all
     * {@link org.eclipse.jface.action.ActionContributionItem}s based on the
     * {@link org.eclipse.jface.action.IAction}s contained in the
     * actions collection
     * @param manager
     * @param actions
     * @param contributionID
     */
    protected void depopulateManager(IContributionManager manager,
            Collection actions) {
        if (actions != null) {
            IContributionItem[] items = manager.getItems();
            for (int i = 0; i < items.length; i++) {
                // Look into SubContributionItems
                IContributionItem contributionItem = items[i];
                while (contributionItem instanceof SubContributionItem) {
                    contributionItem = ((SubContributionItem) contributionItem)
                            .getInnerItem();
                }

                // Delete the ActionContributionItems with matching action.
                if (contributionItem instanceof ActionContributionItem) {
                    IAction action = ((ActionContributionItem) contributionItem)
                            .getAction();
                    if (actions.contains(action)) {
                        manager.remove(contributionItem);
                    }
                }
            }
        }
    }
    
	/**
	 * This populates the specified manager with {@link org.eclipse.jface.action.ActionContributionItem}s
	 * based on the {@link org.eclipse.jface.action.IAction}s contained in the actions collection,
	 * by inserting them before the specified contribution item contributionID.
	 * If ID is null, they are simply added.
     * @param manager
     * @param actions
     * @param contributionID
     */
    protected void populateManager(IContributionManager manager,
            Collection actions, String contributionID) {
        if (actions != null) {
            for (Iterator i = actions.iterator(); i.hasNext();) {
                IAction action = (IAction) i.next();
                if (contributionID != null) {
                    manager.insertBefore(contributionID, action);
                } else {
                    manager.add(action);
                }
            }
        }
    }    

    /**
     * This populates the pop-up menu before it appears
     * @param menuManager
     */
    public void menuAboutToShow(IMenuManager menuManager) {
        super.menuAboutToShow(menuManager);
        MenuManager submenuManager = null;

        submenuManager = new MenuManager(WeaverPlugin.getPlugin().getString(
                "_UI_New_child"));

        populateManager(submenuManager, createChildActions, null);
        menuManager.insertBefore("additions", submenuManager);

        submenuManager = new MenuManager(WeaverPlugin.getPlugin().getString(
                "_UI_New_sibling"));
        populateManager(submenuManager, createSiblingActions, null);
        menuManager.insertBefore("additions", submenuManager);
    }

    /**
     * This inserts global actions before the "additions-end" separator. 
     * @param menuManager
     */
    protected void addGlobalActions(IMenuManager menuManager) {
        menuManager.insertAfter("additions-end", new Separator("ui-actions"));
        menuManager.insertAfter("ui-actions", showPropertiesViewAction);
        menuManager.insertAfter("additions-end", new Separator("ui-actions"));
        menuManager.insertAfter("ui-actions", saveKm3MetamodelAction);
        saveKm3MetamodelAction.setEnabled(false);
        menuManager.insertAfter("additions-end", new Separator("ui-actions"));
        menuManager.insertAfter("additions-end", saveEcoreMetamodelAction);
        menuManager.insertAfter("ui-actions", setParentNameAction);        
        menuManager.add(new Separator("others"));
        super.addGlobalActions(menuManager);  //copy, paste, delete, etc.
    }
    
  
   
    
  		
	/**
	 * This generates a {@link org.eclipse.emf.edit.ui.action.CreateChildAction} for each object in descriptors,
	 * and returns the collection of these actions.
     * @param descriptors
     * @param selection
     * @return
	 */
    protected Collection generateCreateChildActions(Collection descriptors,
            ISelection selection) {
        Collection actions = new ArrayList();
        if (descriptors != null) {
            for (Iterator i = descriptors.iterator(); i.hasNext();) {
                actions.add(new WeaverChildAction(activeEditorPart, selection,
                        i.next(),selectionProvider));
            }
        }
        return actions;
    }
    
    
	/**
	 * This generates a {@link org.eclipse.emf.edit.ui.action.CreateSiblingAction} for each object in descriptors,
	 * and returns the collection of these actions.
     * @param descriptors
     * @param selection
     * @return
     */
    protected Collection generateCreateSiblingActions(Collection descriptors,
            ISelection selection) {
        Collection actions = new ArrayList();
        if (descriptors != null) {
            for (Iterator i = descriptors.iterator(); i.hasNext();) {
                actions.add(new WeaverSiblingAction(activeEditorPart,
                        selection, i.next(),selectionProvider));
            }
        }
        return actions;
    }
}

 