/*******************************************************************************
 * Copyright (c) 2005 INRIA.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Marcos Didonet Del Fabro - initial API and implementation
 *    Guillaume Gueltas
 *******************************************************************************/


package org.eclipse.gmt.weaver.editors.weaving;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.edit.command.CommandParameter;
import org.eclipse.emf.edit.domain.AdapterFactoryEditingDomain;
import org.eclipse.emf.edit.domain.EditingDomain;
import org.eclipse.emf.edit.provider.IEditingDomainItemProvider;
import org.eclipse.emf.edit.ui.dnd.EditingDomainViewerDropAdapter;
import org.eclipse.gmt.weaver.core.WeaverModelManager;
import org.eclipse.gmt.weaver.editors.actions.WeaverChildAction;
import org.eclipse.gmt.weaver.editors.interfaces.IWeavingPanel;
import org.eclipse.gmt.weaver.plugin.util.WeaverInterfaceUtil;
import org.eclipse.gmt.weaver.plugin.util.WeaverUtil;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.swt.SWT;
import org.eclipse.swt.dnd.DND;
import org.eclipse.swt.dnd.DropTargetEvent;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.swt.widgets.MenuItem;

/**
 * Drag and Drop manager. creates a new WLinkEnd if object is dropped in a WLink element or modifies
 * the current WLinkEnd properties 
 */
public class WeavingPanelDndManager extends EditingDomainViewerDropAdapter{

	private IWeavingPanel weavingPanel;
	private WeaverModelManager modelManager = null;
	private EditingDomain editingDomain = null;
	final static int NULL_SELECTION = -1;
	final static int DEFAULT_SELECTION = 0;		
	final static int LINK_END = 1;
	final static int LINK = 2;
	final static int ASSOCIATION = 3;
	final static int ASSOCIATION_END = 4;
	
	private EObject aTargetObj = null;
	private EObject aSource = null;
	private int selectionType;
	
	/**
	 * Creates a new manager of drag and drop for weaving panel
	 * @param panel
	 * @param viewer
	 * @param modelManager
	 * @param editingDomain
	 */
	public WeavingPanelDndManager(IWeavingPanel panel, Viewer treeViewer){
		super(panel.getGlobalEditor().getEditingDomain(),treeViewer);		
		this.editingDomain = panel.getGlobalEditor().getEditingDomain(); 
		this.modelManager = panel.getGlobalEditor().getModelManager();
		this.weavingPanel = panel;
		this.selectionType = NULL_SELECTION;
		this.viewer = treeViewer;
	}
	
	/**
	 * Enables disables drag and drop over specific elements (WLinkEnd and WLink)
	 * @param event
	 */
	public void helper(DropTargetEvent event){		
		selectionType = DEFAULT_SELECTION;
		event.detail=DND.DROP_NONE;
		//gets the source
		source = getDragSource(event);
		aSource = (EObject)source.iterator().next();
		//gets the target
		Object aObj = extractDropTarget( event.item );	        
		if (aObj instanceof EObject){        	
			aTargetObj = (EObject)aObj;
			
			//if the resource is the same, they are consequently in the same window.
			//TODO check this
			if (aSource.eResource().getURI().equals(aTargetObj.eResource().getURI())){				
				if (modelManager.getHelper().isInstanceOfMetaModel(aTargetObj,"WAssociationEnd")){
					if (isValidSourceObj(new String[]{"WLink"}, aSource)){
						event.detail=DND.DROP_LINK;
						selectionType = ASSOCIATION_END;
					}	            				            		
				}
				else if (modelManager.getHelper().isInstanceOfMetaModel(aTargetObj,"WAssociation")){
					if (isValidSourceObj(new String[]{"WLink"}, aSource))	            {
						selectionType = ASSOCIATION;
						event.detail=DND.DROP_LINK;
					}	            				            	
				}
			} else {
				if (modelManager.getHelper().isInstanceOfMetaModel(aTargetObj,"WLinkEnd")){
					selectionType = LINK_END;	
					event.detail=DND.DROP_LINK;
				}		                
				else if (modelManager.getHelper().isInstanceOfMetaModel(aTargetObj,"WLink")){
					selectionType = LINK;
					event.detail=DND.DROP_LINK;
				}		            		            
			}	            	            
		}        
	}      
	
	/**
	 * Verifies if an object can be dropped over another
	 * @param objClassNames
	 * @param eSource
	 * @return
	 */
	public boolean isValidSourceObj(String[] objClassNames, EObject eSource ){
		for (int i = 0; i< objClassNames.length; i ++){
			if (modelManager.getHelper().isInstanceOfMetaModel(eSource, objClassNames[i]))
				return true;
		}
		return false;
	}
	
	/**
	 * the drop action
	 *  @param event
	 */
	public void drop(DropTargetEvent event){
		if (selectionType == DEFAULT_SELECTION )//just enter into the selection code if an object is selected
			super.drop(event);
		else{
			switch (selectionType) {
			case LINK:
				try {
					if (viewer !=null)
						viewer.setSelection(new StructuredSelection(aTargetObj), true);
					selectDropTarget(aSource,aTargetObj,"WLinkEnd",
								WeaverInterfaceUtil.class.getMethod("setNewWLinkEnd",
										new Class [] { EObject.class, EObject.class, 
										String.class, String.class,
										WeaverModelManager.class, EditingDomain.class}));				
				} catch (SecurityException e) {							
					e.printStackTrace();							
				} catch (NoSuchMethodException e) {							
					e.printStackTrace();
				}				
				break;
			case LINK_END:
				WeaverInterfaceUtil.setLinkEndValueProperty(aSource,aTargetObj,
						modelManager,
						editingDomain
				);
				if (viewer !=null)
					viewer.setSelection(new StructuredSelection(aTargetObj), true);
				break;
			case ASSOCIATION:{	  
				try {
					if (viewer !=null)
						viewer.setSelection(new StructuredSelection(aTargetObj), true);
					selectDropTarget(aSource,aTargetObj,"WAssociationEnd",
								WeaverInterfaceUtil.class.getMethod("setNewAssociationEnd",
										new Class [] { EObject.class, EObject.class, 
										String.class, String.class,
										WeaverModelManager.class, EditingDomain.class}));
				} catch (SecurityException e) {							
					e.printStackTrace();							
				} catch (NoSuchMethodException e) {							
					e.printStackTrace();
				}
				break;
			}
			case ASSOCIATION_END:{
				WeaverInterfaceUtil.setAssociatedWLink(aSource,aTargetObj,modelManager,
						editingDomain);
				if (viewer !=null)
					viewer.setSelection(new StructuredSelection(aTargetObj), true);
				break;
			}	        			        		
			}	        		
		}
	}	
	
	/**
	 * Selects the drop target based on the abstract class
	 * @param eSource
	 * @param eTarget
	 * @param eObjTypeDescr
	 * @param method
	 */
	public void selectDropTarget (EObject eSource, EObject eTarget, String eObjTypeDescr, Method method){
		
		AdapterFactoryEditingDomain factory = (AdapterFactoryEditingDomain) editingDomain; 
		IEditingDomainItemProvider adapter = (IEditingDomainItemProvider) factory.getAdapterFactory().adapt(eTarget,IEditingDomainItemProvider.class);

		Collection aList =null;
		if (adapter != null)
			aList = adapter.getNewChildDescriptors(eTarget,editingDomain,null);
		
		List alist = new ArrayList();
		for (Iterator i = aList.iterator(); i.hasNext(); )
			alist.add(i.next());
		createContextDnDMenu(alist,eSource,eTarget,method);
	}
	
	/**
	 * creates a dynamic pop up menu after dragging an element over it if it has more than 1 wlinkend or association end child
	 * @param aList
	 * @param aSource
	 * @param target
	 * @param newElementMethod
	 */
	public void createContextDnDMenu(List aList, EObject aSource, EObject target, Method newElementMethod){
		if (aList.size() == 1){
			try {				
				CommandParameter cmd = (CommandParameter) aList.get(0);	
				newElementMethod.invoke(null,new Object[]{ 
						(EObject)aSource, (EObject)target, 
						cmd.getEValue().eClass().getName(),
						cmd.getEReference().getName(),
						modelManager,
						editingDomain});					
			} catch (IllegalArgumentException e) {
				e.printStackTrace();
			} catch (IllegalAccessException e) {
				e.printStackTrace();
			} catch (InvocationTargetException e) {
				e.printStackTrace();
			}
		}
		else
			if (aList.size() > 1) {//there is more than 1 element, creates a pop up menu
				Menu wElementDndMenu = new Menu(viewer.getControl());
				for (Iterator i = aList.iterator(); i.hasNext();){
					CommandParameter cmd = (CommandParameter) i.next();
					if (((selectionType == ASSOCIATION) && 
							(modelManager.getHelper().isInstanceOfMetaModel(cmd.getEValue(),"WAssociationEnd")))
							|| 					 
							((selectionType != ASSOCIATION) 
									&& modelManager.getHelper().isInstanceOfMetaModel(cmd.getEValue(),"WLinkEnd")))
					{
						WeaverChildAction action = new WeaverChildAction(
								weavingPanel.getGlobalEditor(),viewer.getSelection(),cmd,viewer);													
						MenuItem item  = new MenuItem(wElementDndMenu, SWT.PUSH);
						String wElementTypeDescr = action.getText(); 
						item.setText(wElementTypeDescr);
						if (! cmd.getEReference().isMany() )
							item.setEnabled(target.eGet(cmd.getEReference())==null);
						//puts all the needed data in the menu item
						item.setData(new Object[]{action, target, aSource, cmd.getEValue(), cmd.getEReference().getName()});
						DndSelectionAdapter dndAdapter = new DndSelectionAdapter(viewer);
						item.addSelectionListener(dndAdapter);
						wElementDndMenu.setVisible(true);													
					}						
				}												
			}			
	}
	
	/**
	 * This class adapt the drag and drop
	 */
	class DndSelectionAdapter extends SelectionAdapter {
		Viewer viewer = null;
		
		/**
		 * Creates a drag and drop adapter
		 * @param viewer
		 */
		public DndSelectionAdapter(Viewer viewer){
			this.viewer = viewer;
		}
		
		/* (non-Javadoc)
		 * @see org.eclipse.swt.events.SelectionListener#widgetSelected(org.eclipse.swt.events.SelectionEvent)
		 */
		public void widgetSelected(SelectionEvent e){
			((WeaverChildAction) ((Object[]) e.widget.getData())[0]).run();
			EObject target = (EObject) ((Object[]) e.widget.getData())[1];//target
			EObject source = (EObject) ((Object[]) e.widget.getData())[2];//source
			EObject created = (EObject) ((Object[]) e.widget.getData())[3];//created obj
			String refName = (String) ((Object[]) e.widget.getData())[4];//ref name

			if (selectionType == ASSOCIATION){
				WeaverUtil.setPropertyValue(target,created,refName,editingDomain);
				WeaverInterfaceUtil.setAssociatedWLink(source,created,modelManager,editingDomain);
			}				
			else {
				EObject elementRef = WeaverInterfaceUtil.createGetWElementRef(source, modelManager,editingDomain);
				//set element ref
				WeaverInterfaceUtil.setWLinkEndProperties(created, elementRef,editingDomain);				
			}
		}		
	}
}
