/*******************************************************************************
 * Copyright (c) 2005 INRIA.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Marcos Didonet Del Fabro - initial API and implementation
 *    Guillaume Gueltas
 *******************************************************************************/


package org.eclipse.gmt.weaver.plugin.util;

import java.util.Iterator;

import org.eclipse.core.runtime.IStatus;
import org.eclipse.emf.common.notify.Adapter;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.emf.edit.domain.AdapterFactoryEditingDomain;
import org.eclipse.emf.edit.domain.EditingDomain;
import org.eclipse.gmt.weaver.core.WeaverModelManager;
import org.eclipse.gmt.weaver.exception.WeaverException;
import org.eclipse.gmt.weaver.plugin.WeaverConstant;
import org.eclipse.gmt.weaver.providers.IIdentifierAdapter;
import org.eclipse.gmt.weaver.providers.XMIIdentifierAdapter;

/**
 * Util class for extension panels
 * @since 1.0
 */
public class WeaverInterfaceUtil {
	/**
	 * sets a new associated WLink for the specified WAssociationEnd 
	 * @param wLink the wlink derived instance
	 * @param wAssociationEnd the WAssociationEnd derived instance
	 * @param mngr the model manager
	 * @param domain the editing domain
	 */
	public static void setAssociatedWLink(EObject wLink, EObject wAssociationEnd, WeaverModelManager mngr, EditingDomain domain){
		if (mngr.getHelper().isInstanceOfMetaModel(wLink,"WLink")){
			
			WeaverUtil.setPropertyValue(wAssociationEnd,wLink,"relatedLink", domain);
			WeaverUtil.setPropertyValue(wAssociationEnd,
					wLink.eGet(wLink.eClass().getEStructuralFeature("name")),"name", domain);			
		}
	}	
	
	/**
	 * creates a new WLinkEnd for the selected WLink element
	 * @param aSource the object to be woven
	 * @param target the WLink where the WLinkEnd will be created
	 * @param linkEndTypeDescr the type of WLinkEnd
	 * @param manager
	 * @param editingDomain
	 * @return the created WLinkEnd
	 */
	public static Object setNewWLinkEnd( EObject aSource, EObject target, String linkEndTypeDescr, 
			String referenceName, WeaverModelManager manager, EditingDomain editingDomain ) {
		//creates WelementRef if needed
		EObject elementRef = createGetWElementRef(aSource, manager,editingDomain);
		//creates link end        
		EObject linkEnd = manager.helper.createWeavingElement_32(linkEndTypeDescr,editingDomain);         
		//adds to the link element		
		WeaverUtil.setPropertyValue(target,linkEnd,referenceName,editingDomain);
		//set element ref
		setWLinkEndProperties(linkEnd, elementRef,editingDomain);
		return linkEnd;
	}	
	/**
	 * creates a new WLink for the selected parent element
	 * @param parent it might be a WModel or WLink where the WLink will be created
	 * @param manager
	 * @param editingDomain
	 * @return the created WLink
	 */
	public static Object setNewWLink( EObject parent, 
			 WeaverModelManager manager, EditingDomain editingDomain ) {
		//creates WLink         
		EObject wLink = manager.helper.createWeavingElement_32("Link",editingDomain);         
		//adds to the link element
		if ( manager.getHelper().isInstanceOfMetaModel(parent,"WModel"))
			WeaverUtil.setPropertyValue(parent,wLink ,"ownedElement",editingDomain);				
		return wLink;
	}	
	/**
	 * creates a new WAssociationEnd for the given association and associates the WLink passed as parameter
	 * @param wLink the associated element
	 * @param wAssociation the association
	 * @param associationEndTypeDescr WAssociationEnd type 
	 * @param modelManager
	 * @param editingDomain
	 */
	public static void setNewAssociationEnd( EObject wLink, EObject wAssociation, 
			String associationEndTypeDescr, String referenceName, WeaverModelManager modelManager, EditingDomain editingDomain ) {
		//creates association end        
		EObject wAssociationEnd = 
			modelManager.helper.createWeavingElement_32(associationEndTypeDescr,editingDomain);       
		//adds to the associationEnd reference
		WeaverUtil.setPropertyValue(wAssociation,wAssociationEnd,referenceName,editingDomain);
		//set WLink reference
		setAssociatedWLink(wLink,wAssociationEnd, modelManager, editingDomain);
	}      	
	
	/**
	 * sets the value of name and element feature of a WLinkEnd
	 * @param wLinkEnd 
	 * @param elementRef the WElementRef to be referenced in ref
	 * @param editingDomain
	 */
	public static void setWLinkEndProperties(EObject wLinkEnd, EObject elementRef, EditingDomain editingDomain){
		//set element ref    	
		WeaverUtil.setPropertyValue(wLinkEnd, elementRef, "element",editingDomain);
		//set name gets wref name - prefix"
		EStructuralFeature feature = elementRef.eClass().getEStructuralFeature("name");
		if (feature != null){
			String name = (String)elementRef.eGet(elementRef.eClass().getEStructuralFeature("name"));		
			WeaverUtil.setPropertyValue(wLinkEnd,
					name.substring(WeaverConstant.REF_PREFIX.length()),"name", editingDomain);			
		}
	}
	
	/**
	 * gets (or create) a new WelementRef and generates and ID for it
	 * @param sourceObj the source object
	 * @param manager 
	 * @param editingDomain
	 * @return the created element
	 * @throws WeaverException 
	 */
	public static EObject createGetWElementRef( EObject sourceObj, WeaverModelManager manager, EditingDomain editingDomain) {
		
		EObject wElementRef = null;
		String ID = null;
		//		gets the parent object
		//EObject root = (EObject)EcoreUtil.getRootContainer(sourceObj);
		//gets the associated wmodelref
		
		IIdentifierAdapter adapter = WeaverUtil.getIdentifierAdapter(sourceObj.eResource());
		EObject modelRef = getWModelRef(adapter.getID().toString(), manager);
		if (modelRef == null)
			try {
				throw new WeaverException(IStatus.ERROR,"WModelRef not found");
			} catch (WeaverException e2) {				
				e2.printStackTrace();
			}
		else {				
			//creates a welement ref to calculate the id but not puts it in the resource
			EObject auxWElementRef = null;
			try {				
				String className = WeaverUtil.getAnnotationRefType(modelRef.eClass());				
				if (className == null) {
					//TODO this if is used to keep backward compatibility with 'subsets' keyword. 
					//This should be deleted 
					EReference ownedElementRef = (EReference) modelRef.eClass().getEStructuralFeature("ownedElementRef");
					//creates a welement ref to calculate the id but not puts it in the resource
					auxWElementRef = manager.getHelper().createWeavingElement(
								ownedElementRef.getEType().getName(),editingDomain);
				} else {
					auxWElementRef = manager.getHelper().createWeavingElement(className, editingDomain);					
				}								
				
			} catch (WeaverException e1) {
				e1.printStackTrace();
			}
			
			//gets the reference with the welementref class name		
			ID = getElementID(auxWElementRef, sourceObj,editingDomain);
			//
			wElementRef = searchWElementRef(ID,modelRef,manager,editingDomain);
			if (wElementRef == null) { //if not found WElementRef			
									//put the created wElement ref into the resource
				//sets the reference and the containment with the wmodelref
				setWElementRefReferenceID(ID, sourceObj, auxWElementRef, editingDomain);			
				WeaverUtil.setPropertyValue(modelRef,auxWElementRef,WeaverConstant.ELEMENT_REF_FEATURE, editingDomain);
				wElementRef = auxWElementRef;
			}
		}
		return wElementRef;
	}   
	
	/**
	 * 
	 * @param ID
	 * @param modelRef
	 * @param manager
	 * @param editingDomain
	 * @return
	 */
	
	private static EObject searchWElementRef(String ID, EObject modelRef, WeaverModelManager manager, EditingDomain editingDomain) {
		EObject wElementRef = null;
		for (Iterator i = manager.getWeavingModelRS().getAllContents(); i.hasNext();) {
			Object obj = i.next();
			if (obj instanceof EObject) {
				EObject elemRef = (EObject) obj;
				//search for all ElementRefs				
				if (manager.getHelper().isInstanceOfMetaModel(elemRef,"WElementRef")) {
					//gets the ref attribute
					EObject aModelRef = (EObject) elemRef.eGet(elemRef.eClass().getEStructuralFeature("modelRef"));
					if (aModelRef.equals(modelRef)){
						String referenceStr = (String) elemRef.eGet(elemRef.eClass().getEStructuralFeature("ref"));
						if (referenceStr != null)
							if (referenceStr.equals(ID)) {//if equals returns it
								wElementRef = elemRef;//override the created welement ref for an existing one								
								break;
							}						
					}
				}
			}
		}
		return wElementRef;
	}

	/**
	 * 
	 * @param sourceObj
	 * @param editingDomain
	 * @return
	 */
	private static String getElementID(EObject wElementRef, EObject sourceObj, EditingDomain editingDomain) {
		String ID = "";
		//searchs for an identifier adapter for generating the ID
		AdapterFactoryEditingDomain factory = (AdapterFactoryEditingDomain) editingDomain; 
		Adapter adapter = factory.getAdapterFactory().adapt(wElementRef,IIdentifierAdapter.class);
		if (adapter != null){ //search for object mecanism for creating ids
			if (adapter instanceof IIdentifierAdapter){
				((IIdentifierAdapter)adapter).setTarget(sourceObj);
				ID = (String) ((IIdentifierAdapter)adapter).getID();
			}
			else
				ID = EcoreUtil.getURI(sourceObj).fragment();				
		}
		if (ID == null)
			ID = EcoreUtil.getURI(sourceObj).fragment();		
		return ID;
	}

	/**
	 * sets the ref property with a given ID
	 * @param ID the new generated id
	 * 	@param sourceObj the source woven object
	 * 	@param wElementRef the element
	 * @param editingDomain
	 */
	
	private static void setWElementRefReferenceID(String ID, EObject sourceObj, EObject wElementRef, EditingDomain editingDomain) {
		//set ref
		WeaverUtil.setPropertyValue(wElementRef, ID, "ref", editingDomain);
		//set name
		String name = getNameProperty(sourceObj);
		if (name == null)
			name = ID ;
		EStructuralFeature feature = wElementRef.eClass().getEStructuralFeature("name");
		if (feature != null)
			WeaverUtil.setPropertyValue(wElementRef,WeaverConstant.REF_PREFIX +name,"name", editingDomain);
	}
	/**
	 * tries to get a name property value from an object . if it doees not exist returns its URI
	 * @param sourceObj
	 * @return
	 */
	public static String getNameProperty(EObject sourceObj){
		//tries to get the name properties, if there is not, uses the URI
		EStructuralFeature feature = sourceObj.eClass().getEStructuralFeature("name");
		if (feature == null){//search the Name prop
			feature = sourceObj.eClass().getEStructuralFeature("Name");
			if (feature == null)
				return  EcoreUtil.getURI(sourceObj).fragment() + WeaverConstant.ID_SUFFIX; 
		}
		return  (String)sourceObj.eGet(feature);		
	}
	
	/**
	 * returns the WModelRef that points to the resource passed in the URI
	 * @param uri
	 * @param manager
	 * @return
	 */
	public static EObject getWModelRef(String uri, WeaverModelManager manager){
		Resource res = manager.getWeavingModel();
		
		for (Iterator i = res.getResourceSet().getAllContents();i.hasNext();){
			Object obj = i.next();
			if (obj instanceof EObject){
				EObject eObj = (EObject)obj;
				if (manager.getHelper().isInstanceOfMetaModel(eObj,"WModelRef") ){
					String value =  (String)eObj.eGet( eObj.eClass(). getEStructuralFeature("ref"));
					if (value != null)
						if (value.equals(uri))
							return eObj;                                
				}                
			}
		}
		return null;        
	}    	
	/**
	 * returns the WModelRef that points to the resource passed in the URI
	 * @param uri
	 * @param res
	 * @return
	 */
	public static EObject getWModelRef(String uri, Resource res){
		
		for (Iterator i = res.getResourceSet().getAllContents();i.hasNext();){
			Object obj = i.next();
			if (obj instanceof EObject){
				EObject eObj = (EObject)obj;				
				String value =  (String)eObj.eGet( eObj.eClass(). getEStructuralFeature("ref"));
				if (value != null)
					if (value.equals(uri))
						return eObj;                                
			}
		}
		return null;        
	}    	
			
	/**
	 * sets the element property of the WLinkEnd object associated with the current action
	 * @param aSource the object source with a given URI
	 * @param target the WLinkEnd
	 * @param manager
	 * @param editingDomain
	 */
	public static void setLinkEndValueProperty(EObject aSource, EObject target, WeaverModelManager manager, EditingDomain editingDomain) {    	
		EObject elementRef = createGetWElementRef(aSource, manager,editingDomain);
		setWLinkEndProperties(target,elementRef,editingDomain);		
	}

	/**
	 * return the identifier adapter for generating IDs for a given object. If it does not exists creates a new XMIIdentifierAdapter. 
	 * @param eObj
	 * @return
	 */
	//TODO code replicated in WElementItemProvider and WLinkItemProvider
	public static IIdentifierAdapter createGetIdentifierAdapter(EObject eObj) {
		IIdentifierAdapter adapter = WeaverUtil.getIdentifierAdapter(eObj);
		if (adapter == null){
    		adapter = new XMIIdentifierAdapter(eObj);
    		eObj.eAdapters().add(adapter);        		
		}
		return adapter;
		
	}    
	
	
	
}