/*******************************************************************************
 * Copyright (c) 2005 INRIA.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Marcos Didonet Del Fabro - initial API and implementation
 *    Guillaume Gueltas
 *******************************************************************************/


package org.eclipse.gmt.weaver.plugin.util;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.SequenceInputStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.FileLocator;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.Platform;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EPackage.Registry;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.emf.ecore.xmi.impl.XMIResourceFactoryImpl;
import org.eclipse.gmt.tcs.injector.TCSInjector;
import org.eclipse.gmt.weaver.exception.WeaverException;
import org.eclipse.gmt.weaver.plugin.WeaverConstant;
import org.eclipse.gmt.weaver.plugin.WeaverPlugin;
import org.eclipse.m2m.atl.core.ATLCoreException;
import org.eclipse.m2m.atl.core.IInjector;
import org.eclipse.m2m.atl.core.IModel;
import org.eclipse.m2m.atl.core.IReferenceModel;
import org.eclipse.m2m.atl.core.ModelFactory;
import org.eclipse.m2m.atl.core.emf.EMFModel;
import org.eclipse.m2m.atl.core.launch.ILauncher;
import org.eclipse.m2m.atl.core.service.CoreService;

/**
 * This class contains convenient static methods for managing resource
 * @since 1.0
 */
public class WeaverResourceUtil {
	
	static private String mmName = "KM3";

	/**
	 * returns the full path of a file, given its relative path
	 * @param fileStr
	 * @return
	 */
	public static String getFullPath(String fileStr){
		return ResourcesPlugin.getWorkspace().getRoot().getLocation().toString() + fileStr;
	}
	/**
	 * registers the package in the package registry
	 * @param resource
	 */
	public static Resource registerResource(Resource resource) {
		Registry registry = resource.getResourceSet().getPackageRegistry(); 
		//EPackage.Registry.INSTANCE; 
	    for(Iterator it = WeaverUtil.getElementsByType("EPackage",resource).iterator() ; it.hasNext() ; ) {
	        EPackage pkg = (EPackage)it.next();
	        String nsURI = pkg.getNsURI();
	        if(nsURI == null) {
	            nsURI = pkg.getName();
	            pkg.setNsURI(nsURI);
	        }
	        registry.put(nsURI, pkg);
	        //pkg.setEFactoryInstance(new WFactoryImpl());
	    }
	    return resource;
	}

	
	/**
	 * registers the package in the package registry and associates the instance class name for every primitive type
	 * @param resource
	 */
	public static Resource initResource(Resource resource) {		
		resource = registerResource(resource);	    
	    for(Iterator it = WeaverUtil.getElementsByType("EDataType",resource).iterator(); it.hasNext(); ) {
	        EObject eObj = (EObject)it.next();
	        EStructuralFeature feature;
	        feature = eObj.eClass().getEStructuralFeature("name");
	        String featName = (String)eObj.eGet(feature);
	        String instanceClassName = null;
	        if(featName.equals("Boolean"))
	            instanceClassName = "java.lang.Boolean";
	        else if(featName.equals("Double"))
	            instanceClassName = "java.lang.Double";
	        else if(featName.equals("Float"))
	            instanceClassName = "java.lang.Float";
	        else if(featName.equals("Integer"))
	            instanceClassName = "java.lang.Integer";
	        else if(featName.equals("String"))
	            instanceClassName = "java.lang.String";
	        
	        if (instanceClassName != null) {
	            feature = eObj.eClass().getEStructuralFeature("instanceClassName");
	            eObj.eSet(feature, instanceClassName);
	        }
	    }
	    return resource;
	}
	
	/**
	 * Initializes a resource with a file
	 * The uri corresponds to the to the reference ou the ressource in the resource set
	 * @param file
	 * @param uri
	 * @return
	 * @throws IOException
	 * @throws WeaverException
	 */
	public static ResourceSet initializeResourceSet(String file, URI uri) throws IOException, WeaverException{
		//if there is no path
		int fileExtension = getModelFormat(file);
		
		ResourceSet rs = new ResourceSetImpl();
        rs.setPackageRegistry(EPackage.Registry.INSTANCE);
        rs.getResourceFactoryRegistry().getExtensionToFactoryMap().put("xmi", new XMIResourceFactoryImpl());	    	            
		
//		rs.setPackageRegistry(ASMEMFModel.getResourceSet().getPackageRegistry());
//		rs.setResourceFactoryRegistry(ASMEMFModel.getResourceSet().getResourceFactoryRegistry());			

		Resource res = rs.createResource(uri);
		InputStream stream = UtilFile.getStreamForFile(file);
	
		if(fileExtension == WeaverConstant.KM3){
			// km3 management!!!
			throw new WeaverException(IStatus.ERROR,WeaverPlugin.getPlugin().getString("_UI_error_use_km3"));
		}
		else{
			res.load(stream, Collections.EMPTY_MAP);
		}
		stream.close();	
		return rs;
	}	
	/**
	 * Initializes a resource with a file
	 * The uri corresponds to the to the reference ou the ressource in the resource set
	 * @param rsMM the resource set of the metamodel
	 * @param file
	 * @param uri
	 * @return
	 * @throws IOException
	 * @throws WeaverException
	 */
	public static ResourceSet initializeResourceSet(ResourceSet rsMM, Resource resIfNul, String file, URI uri) throws IOException, WeaverException{
		//if there is no path
		int fileExtension = getModelFormat(file);
		
		ResourceSet rs = new ResourceSetImpl();
		//set the same factory for models and metamodels

//		rs.setPackageRegistry(rsMM.getPackageRegistry());
//		rs.setResourceFactoryRegistry(rsMM.getResourceFactoryRegistry());					

        rs.setPackageRegistry(EPackage.Registry.INSTANCE);
        rs.getResourceFactoryRegistry().getExtensionToFactoryMap().put("xmi", new XMIResourceFactoryImpl());	    	            
		
		Resource res = rs.createResource(uri);
		InputStream stream = UtilFile.getStreamForFile(file);
		
		if(fileExtension == WeaverConstant.KM3){
			// km3 management!!!
			throw new WeaverException(IStatus.ERROR,WeaverPlugin.getPlugin().getString("_UI_error_use_km3"));
		}
		else{
			res.load(stream, Collections.EMPTY_MAP);
		}
		stream.close();	
		return rs;
	}	
	/**
	 * Initializes a resource using a uri
	 * The uri corresponds to the to the reference ou the ressource in the resource set
	 * @param uri
	 * @return
	 * @throws IOException
	 * @throws WeaverException
	 */
	public static ResourceSet initializeResourceSet(List resourceSets, URI uri) throws IOException, WeaverException{
		//Registry registry = 
		for (Iterator i = resourceSets.iterator(); i.hasNext();){
			ResourceSet rs = (ResourceSet) i.next(); //the local resource sets have only 1 resource
			if ( ((Resource)rs.getResources().get(0)).getURI().toString().equals(uri.toString()) )
				return rs;				
		}
		ResourceSet rs = null;
		//EPackage pkg = rs.getPackageRegistry().getEPackage(uri.toString());
		EPackage pkg = (EPackage)EPackage.Registry.INSTANCE.getEPackage(uri.toString());		
		rs = pkg.eResource().getResourceSet();
		return rs;
	}	
	
	/**
	 * Returns the format (extension) of model file
	 * @param filename
	 * @return
	 */
	public static int getModelFormat(String filename){
		String ext = UtilFile.getFileExtension(filename);
		if(ext.equalsIgnoreCase(WeaverPlugin.getPlugin().getString("_UI_extension_1"))) 
			return WeaverConstant.XMI;
		if(ext.equalsIgnoreCase(WeaverPlugin.getPlugin().getString("_UI_extension_2"))) 
			return WeaverConstant.KM3;
		if(ext.equalsIgnoreCase(WeaverPlugin.getPlugin().getString("_UI_extension_3"))) 
			return WeaverConstant.XMI;		
		
		return 0;
	}
	
	/**
	 * Adds stream1 and stream2 and return stream result
	 * @param stream1
	 * @param stream2
	 * @return a SequenceInputStream of stream1 and stream2
	 */
	public static InputStream addStream(InputStream stream1, InputStream stream2){
		return new SequenceInputStream(stream1,stream2);	
	}
	
	/**
	 * returns a unique stream starting from a list of stream
	 * @param inputStreams
	 * @return
	 * @throws WeaverException
	 * @throws WeaverException
	 */
	public static InputStream addListOfStreams(List inputStreams) throws WeaverException {
	    InputStream stream = null;
	    try {
	    	int size = inputStreams.size();
	        for (int i=0;i<size;i++){
	        	// first, we initialize stream
	        	if(i == 0) {
	        		stream = (InputStream)inputStreams.get(i);
	        	}
	            // after first, we add others streams 
	            if(i != 0){
	            	InputStream km3Stream = (InputStream)inputStreams.get(i);
	            	stream = addStream(stream,km3Stream);
	            }
	        }		        
	    } catch (Exception e){
	        e.printStackTrace();
	        throw new WeaverException(IStatus.ERROR,WeaverPlugin.getPlugin().getString("_UI_manager_km3_error"), e);
	    }
	    return stream;
	}
	
	
	/**
	 * Returns an associated xmi resource based on a KM3 file (make transformation)
	 * @param km3Stream
	 * @param ecoreFile
	 * @return
	 * @throws WeaverException
	 */
	public static Resource km3Stream2EcoreResource(InputStream km3Stream, String ecoreFile) throws WeaverException{

		IReferenceModel mmKM3;

		ModelFactory factory;
		IInjector injector;
		ILauncher launcher;
		try {
			factory = CoreService.getModelFactory(WeaverConstant.EMF);
			injector = CoreService.getInjector(WeaverConstant.EMF);
			launcher = CoreService.getLauncher(WeaverConstant.EMF_VM);
			launcher.initialize(Collections.<String, Object> emptyMap());

			//load (inject) KM3
			URL metaModelurl = WeaverResourceUtil.class.getResource("KM3.ecore");
			mmKM3 = factory.newReferenceModel();
			injector.inject(mmKM3, metaModelurl.openStream(), Collections.EMPTY_MAP);				
			String metaModelType = "KM3";	
			// meta model
			IModel km3Model = factory.newModel(mmKM3);
			//Transformation
			URL KM32EMFurl = WeaverResourceUtil.class.getResource("KM32EMF.asm");
			TCSInjector ebnfi = new TCSInjector();

			Map params = new HashMap();
			
			params.put("name", metaModelType);
			params.put("lexerClass", ebnfi.getClass().getClassLoader().loadClass("org.eclipse.gmt.tcs.injector.KM3_ANTLR3Lexer"));
			params.put("parserClass", ebnfi.getClass().getClassLoader().loadClass("org.eclipse.gmt.tcs.injector.KM3_ANTLR3Parser"));
			ebnfi.inject(km3Model, km3Stream, params);
			km3Stream.close();
			
			IModel outModel = factory.newModel(factory.getMetametamodel());
			launcher.addInModel(km3Model, "IN", "KM3");
			launcher.addOutModel(outModel, "OUT", "MOF");

			launcher.launch(ILauncher.RUN_MODE, null, Collections.EMPTY_MAP, KM32EMFurl.openStream());
			return ((EMFModel)outModel).getResource(); //((ASMEMFModel)model).getExtent();
		} catch (Exception e){
			e.printStackTrace();
			throw new WeaverException(IStatus.ERROR,WeaverPlugin.getPlugin().getString("_UI_manager_km3_error"), e);		    
		}	    	    	
	}
		
    /**
     * Returns a List of InputStream with all km3 files extension
     * @param extensionFiles the extensions of metamodel in the extension plugin
     * @param localExtensionFiles 
     * @return the extensions of metamodel in the workspace 
     * @throws WeaverException
     */
    public static List loadKm3ExtensionStreamList(List extensionFiles, List localExtensionFiles) throws WeaverException {	
    	List km3InputStreams = new ArrayList();
    	//the resources from the file system
        for (Iterator i = extensionFiles.iterator(); i.hasNext(); ){
            String extension = (String)i.next();
            InputStream file = null;
            String extensionName = extension;
            String plugin = extensionName.substring(0, extensionName.indexOf("/"));
            String extensionPath = extensionName.substring(extensionName.indexOf("/") + 1);
            URL urlPlatform = Platform.getBundle(plugin).getEntry(extensionPath);
            String km3File = "";
            try {
                URL f = FileLocator.toFileURL(urlPlatform);           
                km3File = f.getPath();
            } catch (IOException e1) {
                e1.printStackTrace();
            }
            try {
                file = UtilFile.getStreamForFile(km3File);
                if (file !=null)
                	km3InputStreams.add(file);
            } catch (FileNotFoundException e){
                e.printStackTrace();
                throw new WeaverException(IStatus.ERROR,WeaverPlugin.getPlugin().getString("_UI_manager_km3_not_found"),e);
            }
        }     	
		//the resources from the workspace
		for (Iterator i = localExtensionFiles.iterator(); i.hasNext(); ){		    
		    String extension = (String)i.next();

			InputStream file = null;           			            
			String extensionName = getFullPath(extension);			        
			try {
			    file = UtilFile.getStreamForFile(extensionName);
			    if (file !=null) 
			    	km3InputStreams.add(file);			            
			} catch (FileNotFoundException e){
			    e.printStackTrace();
			    throw new WeaverException(IStatus.ERROR,WeaverPlugin.getPlugin().getString("_UI_manager_km3_not_found"),e);
			}			            					    			            
		}
		return km3InputStreams;
    }
    
	/**
	 * Validate semantic of weaving metamodel
	 * if error found, throw an exception
	 * @param weavingMetaModel
	 * @throws WeaverException
	 */    
	public static void validateWeavingMetamodelSemantic(Resource weavingMetaModel) throws WeaverException{
        //Each class is extended
		String classList[] = {"WLinkEnd", "WLink", /*"WAssociation", "WAssociationEnd", */"WModel", "WModelRef"};
        for(int i=0;i < classList.length;i++){
            if(WeaverUtil.isExtends(classList[i],weavingMetaModel) == false)
            	throw new WeaverException(IStatus.ERROR,WeaverPlugin.getPlugin().getString("_UI_semantic_mm_error_1",classList[i]));
        }
  	}
	/**
	 * Refreshes the folder of the resource. Used when a new file is created in the folder of the current resource and
	 * when it does not appears immediately. for instance, the (amw.prop) property file 
	 * @param res A resource
	 */
	public static void refreshResourceFolder(Resource res) {
		IWorkspaceRoot root = ResourcesPlugin.getWorkspace().getRoot();
		IResource resource = root.findMember(new Path(
				res.getURI().toString().replaceAll(WeaverConstant.RESOURCE_STRING,"")));
		try {
			resource.getParent().refreshLocal(1, null);
		} catch (CoreException e) {
			e.printStackTrace();
		}
		
	}
	

}
