/*******************************************************************************
 * Copyright (c) 2005 INRIA.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Marcos Didonet Del Fabro - initial API and implementation
 *    Guillaume Gueltas
 *******************************************************************************/


package org.eclipse.gmt.weaver.plugin.util;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.debug.core.DebugPlugin;
import org.eclipse.debug.core.ILaunchConfiguration;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EAnnotation;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.impl.EClassImpl;
import org.eclipse.emf.ecore.impl.EObjectImpl;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.emf.edit.command.AbstractOverrideableCommand;
import org.eclipse.emf.edit.command.AddCommand;
import org.eclipse.emf.edit.command.SetCommand;
import org.eclipse.emf.edit.domain.EditingDomain;
import org.eclipse.gmt.weaver.plugin.WeaverConstant;
import org.eclipse.gmt.weaver.providers.IIdentifierAdapter;

/**
 * This class contains convenient static methods for working with the weaver
 * @author - Marcos Didonet Del Fabro (LINA Labs, INRIA team)<br>
 * @since 1.0
 */
public class WeaverUtil {
		
	/**
	 * Return classifier in the list (for this name)
	 * @param name
	 * @param list
	 * @return
	 */
	private static EClassifier getClassifier(String name, List list){
		EClassifier result = null;
		for (Iterator i=list.iterator();i.hasNext() && result == null;){
			EObject currentElement = (EObject)i.next();
			if (currentElement instanceof EClassifier && ((EClassifier)currentElement).getName().equals(name)) {
				result = (EClass)currentElement;
			}
			else if (currentElement instanceof EPackage) {
				result = WeaverUtil.getClassifier(name, ((EPackage)currentElement).eContents());
			}
		}
		return result;
	}
		 
		
	/**
	 * Returns classifier in the resource (for this name)
	 * @param name
	 * @param resource
	 * @return
	 */
	public static EClassifier getClassifier(String name, Resource  resource){
		return WeaverUtil.getClassifier(name, resource.getContents());
	}	
	
		
    /**
     * returns the first class in the resource with a given name
     * @param modelResource
     * @param name
     * @return
     */
    public static EClass getEClassByName(Resource modelResource, String name){
    	if (modelResource != null){
    		for (Iterator i=modelResource.getResourceSet().getAllContents() ;i.hasNext();){
    		    Object obj = i.next();	    		    
    		    if (obj instanceof EClass && name.equals(  ((EClass)obj).getName()) )
    		        return (EClass)obj;
    		}    		
    	}
		return null;
    }
    
    /**
     * returns all of the EClass for the resource
     * @param modelResource
     * @param name
     * @return
     */
    public static List getAllEClass(Resource modelResource){
		List eClasses = new ArrayList();
    	for (Iterator i=modelResource.getResourceSet().getAllContents() ;i.hasNext();){
		    Object obj = i.next();	    		    
		    if (obj instanceof EClass )
		        eClasses.add( (EClass)obj );
		}
		return eClasses;
    }    

    /**
     * returns the package of a given resource with a name equals to pkgName
     * @param modelResource
     * @return
     */
    public static EPackage getPackage(Resource modelResource, String pkgName){
		for (Iterator i= getAllPackages(modelResource).iterator()  ;i.hasNext();){
		    Object obj = i.next();	    		    
	        if (((EPackage)obj).getName().equals(pkgName))
	            return (EPackage)obj;
		}
		return null;		
    }
    /**
     * returns the total number of elements of a given resource 
     * @param resource
     * @return
     */
    public static int countElements (Resource resource){
    	int size = 0;    	
    	for (Iterator i = resource.getAllContents(); i.hasNext();){
    		size ++;
   			i.next();
    	}    	
    	return size;
    }
    /**
     * returns a list with all packages in a resource
     * @param modelResource
     * @return
     */
    public static List getAllPackages(Resource modelResource){
        List pkgs = new ArrayList();
		for (Iterator i = modelResource.getResourceSet().getAllContents() ;i.hasNext();){
		    Object obj = i.next();	    		    
		    if (obj instanceof EPackage ){
		        pkgs.add(obj);		        		        
		    }
		}
		return pkgs;		        
    }    
    /**
     * returns a EStructural feature of the EObject based on its name 
     * @param eObj
     * @param name
     * @return
     */    
    public static EStructuralFeature getEStructuralFeature(EObject eObj, String name){
        return eObj.eClass().getEStructuralFeature(name);
    }

    /**
     * returns the object (or objects if it is a reference) of a given resource that matches the feature name and value for an object 
     * @param res the resource
     * @param name the object class name 
     * @param featureName the feature name
     * @param featValue the feature value
     * @return
     */
	public static Collection getAllMatchedEObjs(Resource res, String className, String featureName, Object featValue){
	    Collection result = new ArrayList();
	    for(Iterator i = res.getResourceSet().getAllContents();i.hasNext();){
	        Object obj = i.next();
	        if (obj instanceof EObject){
	            EObject eObj = (EObject)obj;
	            if (eObj.eClass().getName().equals(className)){
	                if (  eObj.eClass().getEStructuralFeature(featureName).getEType().getName().equals("EString")){
	                    if ( ((String)eObj.eGet(eObj.eClass().getEStructuralFeature(featureName))).equals(featValue))
	                        result.add(eObj);	                	                    
	                }
	            }	            
	        }
	    }
	    return result;
	}
	
	/**
	 * returns true if this eObj is an instance of the classifier name
	 * @param eObj
	 * @param className
	 * @param classifiers a list of classifiers
	 * @return
	 */
	public static boolean isInstanceOf(EObject eObj, String classifierName, List classifiers){
	    EClassifier classif = WeaverUtil.getClassifier(classifierName,classifiers);
	    if (classif == null )
	        return false;	    
	    return ( 
	            ((EClassImpl)classif).eClass().isSuperTypeOf( ((EObjectImpl)eObj).eClass() ) 
	            || classif.isInstance(eObj)  );	                
	}
	
	/**
	 * returns true if this eObj is an instance of the classifier name
	 * @param eObj
	 * @param classifierName
	 * @param resource (metamodel) a resource with classifiers
	 * @return
	 */
	public static boolean isInstanceOf(EObject eObj, String classifierName, Resource resource){
	    EClassifier classif = WeaverUtil.getClassifier(classifierName,resource);
	    if (classif == null )
	        return false;
	    
	    return ( 
	    		((EClassImpl)classif).eClass().isSuperTypeOf( ((EObjectImpl)eObj).eClass() ) 
	            || classif.isInstance(eObj)  );	                
	}
	
    /**
     * returns the metamodel packages related with this eObj
     * @param eObj
     * @return
     */
    static public List getMetamodelPackages(EObject eObj){
        List elements = new ArrayList();
        EPackage pkg = (EPackage)EcoreUtil.getRootContainer(eObj.eClass());       
        for (Iterator i = WeaverUtil.getAllPackages(pkg.eResource()).iterator(); i.hasNext();){
            elements.add(i.next());
        }
        return elements;        
    }    
	
	/**
	 * returns true if someClass extends of superType(String) in resource
	 * @param someClass the EClass to test
	 * @param superType the name of superclass
	 * @param resource
	 * @return true or false
	 */
	public static boolean isSuperTypeOf(EClass someClass, String superType, Resource resource){
		EClass superTypeEClass = getEClassByName(resource,superType);
		if (superTypeEClass == null)
			return false;
		superTypeEClass.getEAllSuperTypes();
		return superTypeEClass.isSuperTypeOf(someClass);
	}		
	
	/**
	 * returns the list of subtype's name of eClassName in the resource
	 * @param eClassName the name of the eclass
	 * @param resource
	 * @return list of String
	 */
	public static List getSubTypeListOf(String eClassName,Resource resource){
		List result = new ArrayList();
		if (resource != null){
			Iterator eClassesIterator = getAllEClass(resource).iterator();
			while(eClassesIterator.hasNext()){
				EClass eClass = (EClass) eClassesIterator.next();
				EList superTypes = eClass.getEAllSuperTypes();
				for(int i=0;i<superTypes.size();i++){
					EClass superType = (EClass) superTypes.get(i);
					if(isSuperTypeOf(superType,eClassName,resource))
						if(result.contains(eClass.getName())!= true)
								result.add(eClass.getName());
				}
			}			
		}
		return result;
	}
	
	/**
	 * returns true if subClass extends of superClass(String) in resource
	 * @param superClass
	 * @param subClass
	 * @param resource
	 * @return
	 */
	public static boolean isExtends(String superClass, Resource resource){
		if (resource != null){
			EClass superEClass = getEClassByName(resource,superClass);
			Iterator eClassesIterator = getAllEClass(resource).iterator();
			while(eClassesIterator.hasNext()){
				EClass eClass = (EClass) eClassesIterator.next();
				if((superEClass.isSuperTypeOf(eClass))&&(!eClass.getName().equals(superClass)))
					return true;
			}
		}
		return false;
	}

	/**
	 * Gets a set of elements by type
	 * @param type
	 * @param resource
	 * @return
	 */
	public static List getElementsByType(String type, Resource resource) {
	    List ret = new ArrayList();	    
	    for(Iterator i = resource.getAllContents(); i.hasNext(); ) {
	        EObject eo = (EObject)i.next();
	        if (eo.eClass().getName().equals(type)) {
	            ret.add(eo);
	        } 
	    }	    
	    return ret;
	}
	
	
	/**
	 * Gets a set of elements by type, and navigates in the hierarchy
	 * goes up trhough the hierarchy
	 * @param type
	 * @param resource
	 * @return
	 */
	public static List getElementsByTypeHierarchy(String type, Resource resource) {
	    List ret = new ArrayList();	    
	    for(Iterator i = resource.getAllContents(); i.hasNext(); ) {
	        EObject eo = (EObject)i.next();
	        if (eo.eClass().getName().equals(type)) {
	            ret.add(eo);
	        } 
	        else {//gets the hierarchy
	        	for (Iterator j = eo.eClass().getEAllSuperTypes().iterator(); j.hasNext();){
	        		EObject aClass = (EObject)j.next();
	        		Object str = aClass.eGet(aClass.eClass().getEStructuralFeature("name"));
	        		if (str != null){
	        			if ( ((String)str).equals(type)){
	        				ret.add(eo);
	        				break;
	        			}
	        		}
	        	}
	        }
	    }	    
	    return ret;
	}  	

	/**
	 * Returns attribute in resource
	 * @param modelResource
	 * @param name
	 * @return
	 */
	public static EAttribute getEAttributeByName(Resource modelResource, String name){
		for (Iterator i=modelResource.getResourceSet().getAllContents() ;i.hasNext();){
			Object obj = i.next();	    		    
			if (obj instanceof EAttribute && name.equals(((EAttribute)obj).getName()))
				return (EAttribute)obj;
		}
		return null;
	}	 	 
	
	/**
	 * Gets object
	 * @param object
	 * @param name
	 * @return
	 */
	public static Object get(EObject object, String name){
		EStructuralFeature sf = object.eClass().getEStructuralFeature(name);
		Object result = object.eGet(sf);
		return result;
	}
	
	/**
	 * Gets boolean
	 * @param object
	 * @param name
	 * @return
	 */
	public static boolean getBoolean(EObject object, String name){
		return ((Boolean)get(object, name)).booleanValue();
	}	
	
	/**
	 * Gets int
	 * @param object
	 * @param name
	 * @return
	 */
	public static int getInt(EObject object, String name){
		return ((Integer)get(object, name)).intValue();
	}		
	
	/**
	 * Gets String
	 * @param object
	 * @param name
	 * @return
	 */
	public static String getString(EObject object, String name){
		return (String)get(object, name);
	}	
	
	/**
	 * Set object
	 * @param object
	 * @param name
	 * @param newValue
	 */
	public static void set(EObject object, String name, Object newValue) {
		EStructuralFeature sf = object.eClass().getEStructuralFeature(name);
		if (sf == null)
			return;
		if (sf.isMany() && !(newValue instanceof List)) {
			EList list = (EList)object.eGet(sf);
			list.add(newValue);
		} 
		else {
			object.eSet(sf, newValue);
		}		
	}
	
	/**
	 * Set boolean
	 * @param object
	 * @param name
	 * @param newValue
	 */
	public static void set(EObject object, String name, boolean newValue) {
		set(object, name, new Boolean(newValue));
	}
	
	/**
	 * Sets int
	 * @param object
	 * @param name
	 * @param newValue
	 */
	public static void set(EObject object, String name, int newValue) {
		set(object, name, new Integer(newValue));
	}
	
	/**
	 * Sets string
	 * @param object
	 * @param name
	 */
	public static void setName(EObject object, String name){
		set(object, "name", name);
	}
	
	/**
	 * Gets name
	 * @param object
	 * @return
	 */
	public static String getName(EObject object){
		return getString(object, "name");
	}
	
	/**
	 * Sets description
	 * @param object
	 * @param description
	 */
	public static void setDescription(EObject object, String description){
		set(object, "description", description);
	}
	
	/**
	 * Gets description
	 * @param object
	 * @return
	 */
	public static String getDescription(EObject object){
		return getString(object, "description");
	}

	/**
	 * Sets a property value for a given EObject based on the  metamodel 
	 * @param eObj
	 * @param propValue
	 * @param propertyName
	 * @param editingDomain
	 */
	public static void setPropertyValue (EObject eObj, Object objValue, String propertyName, EditingDomain editingDomain){
        AbstractOverrideableCommand addStructuralFeatureCmd = null;
        if (eObj.eClass().getEStructuralFeature(propertyName).isMany()){
    		addStructuralFeatureCmd = new AddCommand(editingDomain,
    				eObj, eObj.eClass().getEStructuralFeature(
    						propertyName), objValue);   
    	}
        else{
    		addStructuralFeatureCmd = new SetCommand(editingDomain,
    				eObj, eObj.eClass().getEStructuralFeature(
    						propertyName), objValue);   
        }        	            
		editingDomain.getCommandStack().execute(addStructuralFeatureCmd);		
	}
	/**
	 * returns the adapter capable of generating an ID for a given object
	 * @param eObj the object 
	 * @return the ID generator adapter
	 */
	public static IIdentifierAdapter getIdentifierAdapter(EObject eObj){
		for (Iterator adapters = eObj.eAdapters().iterator(); adapters.hasNext();){
			Object obj = adapters.next();
			if (obj instanceof IIdentifierAdapter)
				if ( ((IIdentifierAdapter) obj).getTarget() == eObj)
					return (IIdentifierAdapter) obj;			
		}
		return null;
	}
	/**
	 * returns the adapter capable of generating an ID for a resource
	 * @param eObj the object 
	 * @return the ID generator adapter
	 */
	public static IIdentifierAdapter getIdentifierAdapter(Resource resource){
		for (Iterator adapters = resource.eAdapters().iterator(); adapters.hasNext();){
			Object obj = adapters.next();
			if (obj instanceof IIdentifierAdapter)
				return (IIdentifierAdapter) obj;			
		}
		return null;
	}
	/**
	 * returns the name of the class of welementref (saved as an annotation) (if it does not exists, returns ownedElementRef
	 * for the moment every model ref class must have an ownedelementrefannotation
	 * @param eClass
	 * @return
	 */
	public static String getAnnotationRefType (EClass eClass){		
		if (eClass.getEAnnotations().size() > 0) {
			String str = ((EAnnotation)eClass.getEAnnotations().get(0)).getSource();
			str = str.substring(str.indexOf(":")+1, str.length());
			return str;
		}
		else
			return null;
	}
	
	/**
	 * returns the derived values for the feature of a class
	 * @param resource
	 * @param eClass
	 * @param sFeature
	 * @return
	 */
	public static List getDerivedReferences(Resource resource, String eClass,String sFeature){
		List result= new ArrayList();
		
		EClass model = WeaverUtil.getEClassByName(resource,eClass);
		EStructuralFeature feature = model.getEStructuralFeature(sFeature);
		
		if (feature != null){
			if (feature.isDerived()){
				String[] featureName = ((EAnnotation)feature.getEAnnotations().get(0)).getSource().split(",");
				List references = model.getEReferences();
				for (int i = 0; i < featureName.length; i++){
					for (Iterator refs = references.iterator(); refs.hasNext();){
						EReference subsetOf = (EReference) refs.next();
						if(subsetOf.getName().equals(featureName[i])){
							result.add(subsetOf);
						}
					}	
				}
			}			
		}	
		return result;
	}	
	/**
	 * returns all the references to WModelREf from o given class
	 * @param resource
	 * @param eClass
	 * @param sFeature
	 * @return
	 */
	public static List getWModelRefReferences(Resource resource, String className){
		List result= new ArrayList();
		
		EClass modelClass = WeaverUtil.getEClassByName(resource, className);
		
		EList references = modelClass.getEAllReferences();				
		for (Iterator refs = references.iterator(); refs.hasNext();){
			EReference ref = (EReference) refs.next();
			if (WeaverUtil.isSuperTypeOf(ref.getEReferenceType(),"WModelRef",resource)){
				result.add(ref);
			}
		}	
		return result;
	}
	/**
	 * gets the property file complete path
	 * @param file
	 * @return
	 */
	public static IPath getPropertyFilePath (IFile file) {
		//takes the current path	
		IPath path = file.getLocation().removeFileExtension();
		IPath newPath = path.addFileExtension(WeaverConstant.PROPERTY_FILE_EXT);
		//project.getLocation() getWorkingLocation(WeaverPlugin.getPlugin().getString("pluginID"));
		return newPath;
	}
	/**
	 * returns a launch configuration of a transformation with the given launch configuration file name
	 * @param fName launch configuration file name 
	 * @return
	 */
    public static ILaunchConfiguration getConfig(String fName){
    	ILaunchConfiguration aConfig = null;
		try {
			ILaunchConfiguration list[];
			list = DebugPlugin.getDefault().getLaunchManager().getLaunchConfigurations();
	    	for (int j=0; j < list.length;j++){	    		
	    		if (list[j].getFile()!=null) {
	    			if (list[j].getFile().getName().equals(fName)){
	    				aConfig = list[j];
	    			}
	    		}
	    	}    	
		} catch (CoreException e) {
			e.printStackTrace();
		}
		return aConfig;
    }

}
