/*******************************************************************************
 * Copyright (c) 2005 INRIA.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
  *
 * Contributors:
 *    Marcos Didonet Del Fabro - initial API and implementation
 *    Guillaume Gueltas
*
 *******************************************************************************/



package org.eclipse.gmt.weaver.wizards;

import java.io.File;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;

import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IExtension;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.QualifiedName;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.gmt.weaver.core.WeaverExtensionManager;
import org.eclipse.gmt.weaver.core.WeaverModelManager;
import org.eclipse.gmt.weaver.core.WeaverXMLMetadata;
import org.eclipse.gmt.weaver.core.data.WovenModelDescriptor;
import org.eclipse.gmt.weaver.exception.WeaverException;
import org.eclipse.gmt.weaver.plugin.WeaverConstant;
import org.eclipse.gmt.weaver.plugin.WeaverPlugin;
import org.eclipse.gmt.weaver.plugin.util.UtilFile;
import org.eclipse.gmt.weaver.plugin.util.WeaverResourceUtil;
import org.eclipse.gmt.weaver.plugin.util.WeaverUtil;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.operation.IRunnableWithProgress;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.wizard.Wizard;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IEditorReference;
import org.eclipse.ui.IFileEditorInput;
import org.eclipse.ui.INewWizard;
import org.eclipse.ui.IWorkbench;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.IWorkbenchWizard;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.WorkbenchException;
import org.eclipse.ui.ide.IDE;
import org.eclipse.ui.internal.ide.DialogUtil;
import org.eclipse.ui.internal.ide.IDEWorkbenchMessages;


/**
 * Wizard of the weaver
 * @since 1.0
 */
public class WeaverWizard extends Wizard implements INewWizard {

	protected WeaverFirstWizardPage pageOne;
	protected WeaverSecondWizardPage pageTwo;	
	protected WeaverThirdWizardPage pageThree;
	protected List activeEditors;//IEditorReference
	private IStructuredSelection selection;
	private Collection metamodelExtensionNames;
	private List wovenPanelExtensionNames = new ArrayList();
	private String[] weavingPanelExtensionNames;
	private List wModelNames;
	protected List wModelRefNames;
	private boolean error= false;
	private boolean overwriteFile = true;
	private String metamodelFile;
	
	// when metamodel extension are selected
	private Resource weavingMetamodel=null;
	
	/**
	 * Creates a new weaver wizard 
	 */
	public WeaverWizard() {
		super();
		//setNeedsProgressMonitor(true);
	}
	
	/**
	 * Adding the pages to the wizard.
	 */
	public void addPages() {
		if(!error){
			
			// first Page
	        pageOne = new WeaverFirstWizardPage(this, "one", selection);
	        pageOne.setTitle(WeaverPlugin.getPlugin().getString("_UI_wizard_first_title"));
	        pageOne.setDescription(WeaverPlugin.getPlugin().getString("_UI_wizard_first_description"));
	        addPage(pageOne);
	        
			// Second Page 
	        pageTwo = new WeaverSecondWizardPage(this,"two", selection);
	        pageTwo.setTitle(WeaverPlugin.getPlugin().getString("_UI_wizard_second_title"));
	        pageTwo.setDescription(WeaverPlugin.getPlugin().getString("_UI_wizard_second_description"));
	        addPage(pageTwo);
	        
			// Second Page 
	        pageThree = new WeaverThirdWizardPage(this,"three", selection);
	        pageThree.setTitle(WeaverPlugin.getPlugin().getString("_UI_wizard_third_title"));
	        pageThree.setDescription(WeaverPlugin.getPlugin().getString("_UI_wizard_third_description"));
	        addPage(pageThree);	       
		}
	}
	
	/**
	 * This method is called when 'Finish' button is pressed in
	 * the wizard. We will create an operation and run it
	 * using wizard as execution context.
	 * @return
	 */
	public boolean performFinish() {
		final List mmPluginExtensions = (List)pageOne.getPluginMmExtensions();
		final List mmWorkspaceExtensions = (List)pageOne.getWorkspaceMmExtensions();
		
		final String projectName = pageTwo.getProjectName();
		final String existingWeavingModel = pageTwo.getExistingModelFileName();
		final String newWeavingModelRelativePath = pageTwo.getNewModelFileName();
		final String weawingPanel = pageTwo.getEditorExtensionSelected();
		final String wmodel = pageTwo.getWmodelSelected();
		
		final List models = pageThree.getModels();
		
		final boolean isNewWeavingModel = pageTwo.isNewWeavingModel();

		IRunnableWithProgress op = new IRunnableWithProgress() {
			public void run(IProgressMonitor monitor) throws InvocationTargetException, InterruptedException {
				try {
					doFinish(
							mmPluginExtensions,
							mmWorkspaceExtensions,
							projectName,
							existingWeavingModel,
							newWeavingModelRelativePath,
							weawingPanel,
							wmodel,
							models,
							isNewWeavingModel,
							monitor);
				} catch (CoreException e) {
					throw new InvocationTargetException(e);
				} finally {
					monitor.done();
				}
			}
		};
		
		try {
			getContainer().run(true, false, op);
		} catch (InterruptedException e) {
			return false;
		} catch (InvocationTargetException e) {
			e.printStackTrace();
			Throwable realException = e.getTargetException();
			MessageDialog.openError(getShell(), WeaverPlugin.getPlugin().getString("_UI_error_label"), realException.getMessage());
			return false;
		}
		return true;
	}

	/**
	 * try to finish 
	 * @param isNewWeavingModel true if model/metamodel is new
	 * @param models the list of wovenModelDescriptor for this model/metamodel
	 * @param wmodel the wmodel choosen
	 * @param weawingPanel weaving panel extension id
	 * @param newWeavingModelRelativePath weaving model file name
	 * @param existingWeavingModel weaving model file path
	 * @param projectName project name
	 * @param mmPluginExtensions list of extensions used
	 * @param mmWorkspaceExtensions list of local extensions used
	 * @param monitor progress monitor
	 * @throws CoreException
	 * @throws InterruptedException 
	 */
	private void doFinish(List mmPluginExtensions, List mmWorkspaceExtensions, String projectName, 
			String existingWeavingModel, String newWeavingModelRelativePath, 
			String weawingPanel, String wmodel, 
			List models,
			boolean isNewWeavingModel, IProgressMonitor monitor) throws CoreException, InterruptedException {		
		String metamodelUri = "";
		// Work 1 : Valid inputs
		monitor.beginTask(WeaverPlugin.getPlugin().getString("_UI_wizard_task_1"), 3);
		IWorkspaceRoot root = ResourcesPlugin.getWorkspace().getRoot();
		IResource resource = root.findMember(new Path(projectName));

		if ((resource==null) || (!resource.exists()) || !(resource instanceof IContainer))
			throw new WeaverException(IStatus.ERROR, WeaverPlugin.getPlugin().getString("_UI_container")
					+" "+WeaverPlugin.getPlugin().getString("_UI_wizard_test_project_neg",projectName));
		IContainer container = (IContainer) resource;
					
		// Test if existing weaving file really exits
		String existingWeavingModelPath="";
		if(!isNewWeavingModel){
			existingWeavingModelPath = ResourcesPlugin.getWorkspace().getRoot().getLocation()+
				container.getFile(new Path(existingWeavingModel)).getFullPath().toString();
			if(!UtilFile.testIfFileExist(existingWeavingModelPath)){
				throw new WeaverException(IStatus.ERROR,WeaverPlugin.getPlugin().getString("_UI_weaving_model")
						+" "+WeaverPlugin.getPlugin().getString("_UI_wizard_test_file_neg",existingWeavingModel));
			}
		}
		
		List modelInfos = new ArrayList();
		final int numberOfModels = models.size();
		for(int i = 0; i < numberOfModels; i++){
			// Descriptor
			WovenModelDescriptor desc = (WovenModelDescriptor) models.get(i);
			// File path
			String aModelfilePath = null;
			String aMetaModelfilePath = null;
			
			// test file
		 	aMetaModelfilePath = desc.getMetamodelFile();
		 			 	
		 	//enables metamodel by uri (must enter http:// )
		 	if (aMetaModelfilePath.indexOf("http") > -1) {
		 		metamodelUri = aMetaModelfilePath;
		 		desc.setMetamodelFile(metamodelUri);
		 	} else		 		
		 		if(!UtilFile.testIfFileExist(root.getLocation() + aMetaModelfilePath))
		 			throw new WeaverException(IStatus.ERROR,WeaverPlugin.getPlugin().getString("_UI_meta_model")+" "
		 					+ WeaverPlugin.getPlugin().getString("_UI_wizard_test_file_neg",aMetaModelfilePath));
		 		else{
		 			desc.setMetamodelFile(aMetaModelfilePath);
		 		}
		 	
		 	//set the model path
		 	if (desc.isModelExists()) {
		 		aModelfilePath = desc.getModelFile();
			 	if(aModelfilePath.length() != 0){
					if(!UtilFile.testIfFileExist(root.getLocation() + aModelfilePath))
						throw new WeaverException(IStatus.ERROR,WeaverPlugin.getPlugin().getString("_UI_model")+" "
							+WeaverPlugin.getPlugin().getString("_UI_wizard_test_file_neg",aModelfilePath));
					else 
						desc.setModelFile(aModelfilePath);
			 	}
		 	}
			
			// update descriptor
			desc.setId(Integer.toString(i));			
			if((aModelfilePath != null)&&(aModelfilePath.length() != 0))
				desc.setModelFile(aModelfilePath);
			else 
				desc.setModelFile("null");

			modelInfos.add(desc);
		}
		monitor.worked(1);	

		// work 2 : Initialize weaving model file
		final IFile file;
		if(isNewWeavingModel){
			monitor.setTaskName(WeaverPlugin.getPlugin().getString("_UI_wizard_task_2",newWeavingModelRelativePath));
			file = initializeNewFile(container, newWeavingModelRelativePath,monitor);
		}
		else{		
			monitor.setTaskName(WeaverPlugin.getPlugin().getString("_UI_wizard_task_2",existingWeavingModel));
			file = container.getFile(new Path(existingWeavingModel)); 
				
				//initializeExistingFile(container,existingWeavingModelPath,monitor);
		}

		// if file already open
		int numberOfFile= activeEditors.size();
		for (int i=0; i<numberOfFile;i++){
			IEditorReference ref = (IEditorReference) activeEditors.get(i);
			IEditorPart part = (IEditorPart) ref.getPart(false);
			if(part!=null){
				IFileEditorInput input = (IFileEditorInput) part.getEditorInput();
				if(input != null){
					IFile openedFile = input.getFile();
					if((openedFile !=null) &&(openedFile.equals(file))){
						throw new WeaverException(IStatus.ERROR,WeaverPlugin.getPlugin().getString("_UI_wizard_test_file_open",file.getFullPath()));
					}
				}
			}

		}

		monitor.worked(1);
		
		// Put path in config file	
		IProject project = container.getProject();
		
		File propertiesFile = WeaverUtil.getPropertyFilePath(file).toFile();
		try {
			// Config file
			WeaverXMLMetadata xml = new WeaverXMLMetadata(propertiesFile);
			xml.createXMLConfigFile(modelInfos,file.getFullPath().toString(),weawingPanel,wmodel,
					mmPluginExtensions, mmWorkspaceExtensions, metamodelFile );
			
			// metadata persistent property for tell if new model
			QualifiedName qualifiedNameNewModelWM = new QualifiedName(file.getFullPath().toString(),WeaverPlugin.getPlugin().getString("_UI_weaving_model"));
			project.setPersistentProperty(qualifiedNameNewModelWM, Boolean.toString(isNewWeavingModel));
		} catch (Exception e1) {
			e1.printStackTrace(System.err);
			throw new WeaverException(IStatus.ERROR,WeaverPlugin.getPlugin().getString("_UI_wizard_write_error")+e1.getMessage());
		}
		// Work 3 : 
		monitor.setTaskName(WeaverPlugin.getPlugin().getString("_UI_wizard_task_3"));
		Shell shell = getShell();
		Display display = shell.getDisplay();
		display.asyncExec(new Runnable() {
			public void run() {
				// Init of weaving perspective
				//site.getPage().getWorkbenchWindow().getWorkbench().addWindowListener(windowListener);
			    IWorkbench workbench = PlatformUI.getWorkbench();
			    IWorkbenchWindow window = workbench.getActiveWorkbenchWindow();
				IWorkbenchPage page = window.getActivePage();
				try {
					//uncomment this line to always open the weaver perspective while creating a weaving model
					//workbench.showPerspective("org.eclipse.gmt.weaver.perspective.weaverPerspectiveID", workbench.getActiveWorkbenchWindow());
					IDE.openEditor(page, file, "org.eclipse.gmt.weaver.globalWeaverEditorID",true);
				}
				catch (PartInitException e) {
					e.printStackTrace();
					DialogUtil.openError(
							page.getWorkbenchWindow().getShell(),
							IDEWorkbenchMessages.OpenWithMenu_dialogTitle, //$NON-NLS-1$
							e.getMessage(),
							(PartInitException) e);
				}
				catch (WorkbenchException e) {
					e.printStackTrace();
					MessageDialog.openError(									
							window.getShell(),
							IDEWorkbenchMessages.OpenWithMenu_dialogTitle, //$NON-NLS-1$
							e.getMessage());
				}
			}
		});
		monitor.worked(1);
	}

	/**
	 * Initializes the new file in the container (project)
	 * @param container the container
	 * @param filePathName if new model : name of file
	 * @param monitor the progress monitor
	 * @throws CoreException
	 */
	private IFile initializeNewFile(IContainer container, String filePathName, IProgressMonitor monitor) throws CoreException {
		IFile file = container.getFile(new Path(filePathName));

		if (file.exists()) {
			throw new WeaverException(IStatus.ERROR,WeaverPlugin.getPlugin().getString("_UI_wizard_test_file_pos",filePathName)+" in the container");
		} 
		else {
			file.create(null, true, monitor);
		}
		return file;
	}
		
	/**
	 * We will accept the selection in the workbench to see if
	 * we can initialize from it.
	 * @see IWorkbenchWizard#init(IWorkbench, IStructuredSelection)
	 */
	public void init(IWorkbench workbench, IStructuredSelection selection) {
		this.selection = selection;
		
		// Get the file opened in the workspace
		activeEditors = new ArrayList();
		IEditorReference ref[] = workbench.getActiveWorkbenchWindow().getActivePage().getEditorReferences();
		activeEditors = Arrays.asList(ref);
		// Initialyse extensions
		initExtensions();
	}
	
	/**
	 * Initializes extensions of the weaver
	 */
	private void initExtensions() {
		// Metamodel Extension
		IExtension extWovenPanel[] = WeaverExtensionManager.getAllExtensions(WeaverConstant.WOVEN_PANEL_EXT_ID);
		IExtension extWeavingPanel[] = WeaverExtensionManager.getAllExtensions(WeaverConstant.WEAVING_PANEL_EXT_ID);
		
		int lenghtExtWovenPanel = extWovenPanel.length;
		int lenghtExtWeavingPanel = extWeavingPanel.length;
		
		WeaverExtensionManager manager = new WeaverExtensionManager(null);
		metamodelExtensionNames = manager.getPluginMetamodelPath();
		
		//wovenPanelExtensionNames = new String[lenghtExtWovenPanel];
		for(int i=0; i<lenghtExtWovenPanel; i++){
			wovenPanelExtensionNames.add(new String (
					extWovenPanel[i].getSimpleIdentifier()));
		}		

		weavingPanelExtensionNames = new String[lenghtExtWeavingPanel];
		for(int i=0;i<lenghtExtWeavingPanel;i++){
			weavingPanelExtensionNames[i] = extWeavingPanel[i].getSimpleIdentifier();
		}	
		
		error = false;
		
		if(lenghtExtWeavingPanel == 0){
			MessageDialog.openWarning(getShell(),WeaverPlugin.getPlugin().getString("_UI_warning_label"),WeaverPlugin.getPlugin().getString("_UI_wizard_warning_extension_4"));
			error = true;
		}		
		if(lenghtExtWovenPanel == 0){
			MessageDialog.openWarning(getShell(),WeaverPlugin.getPlugin().getString("_UI_warning_label"),WeaverPlugin.getPlugin().getString("_UI_wizard_warning_extension_5"));
			error = true;
		}
	}
	
	/**
	 * Initializes wizard pages :<br>
	 *  - WMODEL combo with the WMODEL name found in metamodel extensions selected<br>
	 *  - WMODELREF combo with the WMODELREF name found in metamodel extensions selected<br>
	 *  - Model name with name of association in metamodel extensions selected<br>
	 * @param extensionsSelected
	 * @param monitor
	 * @throws WeaverException
	 */
	public void initializeWeavingResource(Collection extensionsSelected, Collection localExtensionsSelected,
			IProgressMonitor monitor, boolean isEcore, String ecoreMetamodel) throws WeaverException{
		int metamodelType;
		if (isEcore) 
			metamodelType = WeaverConstant.XMI;
		else
			metamodelType = WeaverConstant.KM3;
		
		metamodelFile = ecoreMetamodel;
		
		monitor.beginTask("Add metamodel extension", 1);
		List arrayWModel = new ArrayList();	
		
		// Initialize a manager for wizard
		WeaverModelManager modelManager = new WeaverModelManager((List)extensionsSelected,
				(List)localExtensionsSelected, metamodelType, ecoreMetamodel);
		weavingMetamodel = modelManager.getWeavingMetamodel();
		
		// initialize wmodel control
		arrayWModel = WeaverUtil.getSubTypeListOf("WModel",weavingMetamodel);
		if(arrayWModel.size()<1){
			throw new WeaverException(IStatus.ERROR,WeaverPlugin.getPlugin().getString("_UI_wizard_error_wmodel"));
		}else {
			pageTwo.initWmodelControl(arrayWModel);
		}
		
		monitor.worked(1);
	}
	
	/**
	 * Initializes wizard pages :<br>
	 *  - WMODEL combo with the WMODEL name found in metamodel extensions selected<br>
	 *  - WMODELREF combo with the WMODELREF name found in metamodel extensions selected<br>
	 *  - Model name with name of association in metamodel extensions selected<br>
	 * @param extensionsSelected
	 * @param monitor
	 * @throws WeaverException
	 */
	public void initializeWModelRef(String wModel) throws WeaverException{

		List wModelReference = WeaverUtil.getWModelRefReferences(weavingMetamodel,wModel); 
			
		if (wModelReference.size() == 0){
			String reference = "wovenModel";
			throw new WeaverException(IStatus.ERROR,WeaverPlugin.getPlugin().getString("_UI_semantic_mm_error_2",reference));	
		}
		else {
			pageThree.initControls(wModelReference, weavingMetamodel);
		}
	}	
	
	/**
	 * Validates if model is in conformity with metamodel
	 * @param modelFile
	 * @param metamodelFile
	 * @return
	 * @throws WeaverException
	 */
	public boolean validateModel(String modelFile, String metamodelFile) throws WeaverException{
		ResourceSet rs = null;
		URI uri = null;
		try {
			// initialize resource
			uri = URI.createFileURI(UtilFile.getFileNameFromPath(metamodelFile));
			rs = WeaverResourceUtil.initializeResourceSet(metamodelFile,uri);
			// register the mm
			WeaverResourceUtil.registerResource(rs.getResource(uri,false));
			// initialize resource
			//URI modelUri = URI.createFileURI(UtilFile.getFileNameFromPath(modelFile));
		} catch (Exception e) {
			//if there isn't exception, result is ok
			throw new WeaverException(IStatus.ERROR,WeaverPlugin.getPlugin().getString("_UI_wizard_error_validate_metamodel",
					UtilFile.getFileNameFromPath(modelFile),UtilFile.getFileNameFromPath(metamodelFile)),
					e);
		}	
		return  true;
	}

	/**
	 * @return Returns the pageOne.
	 */
	public WeaverFirstWizardPage getPageOne() {
		return pageOne;
	}
	/**
	 * @return Returns the pageThree.
	 */
	public WeaverThirdWizardPage getPageThree() {
		return pageThree;
	}
	/**
	 * @return Returns the pageTwo.
	 */
	public WeaverSecondWizardPage getPageTwo() {
		return pageTwo;
	}	
	
	/**
	 * @return Returns the activeEditor.
	 */
	protected List getActiveEditors() {
		return activeEditors;
	}
	/**
	 * @return Returns the metamodelExtensionNames.
	 */
	protected Collection getMetamodelExtensionNames() {
		return metamodelExtensionNames;
	}

	/**
	 * @return Returns the wovenPanelExtensionNames.
	 */
	protected List getWovenPanelExtensionNames() {
		return wovenPanelExtensionNames;
	}
	/**
	 * @return Returns the weavingPanelExtensionNames.
	 */
	protected String[] getWeavingPanelExtensionNames() {
		return weavingPanelExtensionNames;
	}
	/**
	 * @return Returns the wModelNames.
	 */
	public List getWModelNames() {
		return wModelNames;
	}
	/**
	 * @param modelNames The wModelNames to set.
	 */
	public void setWModelNames(List modelNames) {
		wModelNames = modelNames;
	}
	/**
	 * @return Returns the overwriteFile.
	 */
	public boolean isOverwriteFile() {
		return overwriteFile;
	}
	/**
	 * @param overwriteFile The overwriteFile to set.
	 */
	public void setOverwriteFile(boolean overwriteFile) {
		this.overwriteFile = overwriteFile;
	}
	/**
	 * returns weaving metamodel
	 * when metamodel extensions are selected
	 * @return Returns the weavingMetamodel.
	 */
	public Resource getWeavingMetamodel() {
		return weavingMetamodel;
	}
	/**
	 * @param weavingMetamodel The weavingMetamodel to set.
	 */
	public void setWeavingMetamodel(Resource weavingMetamodel) {
		this.weavingMetamodel = weavingMetamodel;
	}
	
}